from ziplime.finance.controls.trading_control import TradingControl


class AssetDateBounds(TradingControl):
    """TradingControl representing a prohibition against ordering an asset before
    its start_date, or after its end_date.
    """

    def __init__(self, on_error):
        super(AssetDateBounds, self).__init__(on_error)

    def validate(self, asset, amount, portfolio, algo_datetime, algo_current_data):
        """Fail if the algo has passed this Asset's end_date, or before the
        Asset's start date.
        """
        # If the order is for 0 shares, then silently pass through.
        if amount == 0:
            return

        normalized_algo_dt = algo_datetime.normalize().tz_localize(None)

        # Fail if the algo is before this Asset's start_date
        if asset.start_date:
            normalized_start = asset.start_date.normalize()
            if normalized_algo_dt < normalized_start:
                metadata = {"asset_start_date": normalized_start}
                self.handle_violation(asset, amount, algo_datetime, metadata=metadata)
        # Fail if the algo has passed this Asset's end_date
        if asset.end_date:
            normalized_end = asset.end_date.normalize()
            if normalized_algo_dt > normalized_end:
                metadata = {"asset_end_date": normalized_end}
                self.handle_violation(asset, amount, algo_datetime, metadata=metadata)



"""
Filters that apply smoothing operations on other filters.

These are generally useful for controlling/minimizing turnover on existing
Filters.
"""

from .filter import CustomFilter


class All(CustomFilter):
    """
    A Filter requiring that assets produce True for ``window_length``
    consecutive days.

    **Default Inputs:** None

    **Default Window Length:** None
    """

    def compute(self, today, assets, out, arg):
        out[:] = arg.sum(axis=0) == self.window_length


class Any(CustomFilter):
    """
    A Filter requiring that assets produce True for at least one day in the
    last ``window_length`` days.

    **Default Inputs:** None

    **Default Window Length:** None
    """

    def compute(self, today, assets, out, arg):
        out[:] = arg.sum(axis=0) > 0


class AtLeastN(CustomFilter):
    """
    A Filter requiring that assets produce True for at least N days in the
    last ``window_length`` days.

    **Default Inputs:** None

    **Default Window Length:** None
    """

    params = ("N",)

    def compute(self, today, assets, out, arg, N):
        out[:] = arg.sum(axis=0) >= N



import asyncio
import datetime
import logging

import structlog
from exchange_calendars import ExchangeCalendar
from lime_trader import AsyncLimeClient, LimeClient
from lime_trader.models.market import Period

import polars as pl

from ziplime.assets.entities.asset import Asset
from ziplime.data.services.data_bundle_source import DataBundleSource


class LimeTraderSdkDataSource(DataBundleSource):
    def __init__(self, lime_sdk_credentials_file: str | None,
                 trading_calendar: ExchangeCalendar
                 ):
        super().__init__()
        self._lime_sdk_credentials_file = lime_sdk_credentials_file
        self.trading_calendar = trading_calendar
        self._logger = structlog.get_logger(__name__)
        if lime_sdk_credentials_file is None:
            self._lime_sdk_client = AsyncLimeClient.from_env(logger=self._logger)
            self._lime_sdk_client_sync = LimeClient.from_env(logger=self._logger)
        else:
            self._lime_sdk_client = AsyncLimeClient.from_file(lime_sdk_credentials_file, logger=self._logger)
            self._lime_sdk_client_sync = LimeClient.from_file(lime_sdk_credentials_file, logger=self._logger)

    def get_spot_value(self, assets: frozenset[Asset], fields: frozenset[str], dt, data_frequency,
                       exchange_name: str,
                       exchange_country: str,
                       ) -> pl.DataFrame:

        symbols = [asset.get_symbol_by_exchange(exchange_name=exchange_name) for asset in assets]
        quotes = self._lime_sdk_client_sync.market.get_current_quotes(symbols=symbols)

        cols = {"open": [], "close": [], "price": [], "high": [], "low": [], "volume": [], "date": [], "exchange": [],
                "symbol": [], "exchange_country": []}

        for result in quotes:
            cols["open"].append(result.open)
            cols["close"].append(result.close)
            cols["price"].append(result.close)
            cols["high"].append(result.high)
            cols["low"].append(result.low)
            cols["volume"].append(result.volume)
            cols["date"].append(result.date.astimezone(self.trading_calendar.tz))
            cols["exchange"].append(exchange_name)
            cols["exchange_country"].append(exchange_country)
            cols["symbol"].append(result.symbol)
        df = pl.DataFrame(cols, schema=[("open", pl.Float64()), ("close", pl.Float64()),
                                        ("price", pl.Float64()),
                                        ("high", pl.Float64()), ("low", pl.Float64()),
                                        ("volume", pl.Float64()),
                                        ("date", pl.Datetime), ("exchange", pl.String),
                                        ("exchange_country", pl.String), ("symbol", pl.String)
                                        ])
        return df.select(pl.col(col) for col in fields)

    def get_data_sync(self, symbols: list[str],
                      frequency: datetime.timedelta,
                      date_from: datetime.datetime,
                      date_to: datetime.datetime,
                      ) -> pl.DataFrame:

        results = [self._lime_sdk_client_sync.market.get_quotes_history(
            symbol=symbol, period=self._frequency_to_period(frequency=frequency), from_date=date_from,
            to_date=date_to
        ) for symbol in symbols]

        cols = {"open": [], "close": [], "price": [], "high": [], "low": [], "volume": [], "date": [], "exchange": [],
                "symbol": [], "exchange_country": []}
        for results, symbol in zip(results, symbols):
            for result in results:
                cols["open"].append(result.open)
                cols["close"].append(result.close)
                cols["price"].append(result.close)
                cols["high"].append(result.high)
                cols["low"].append(result.low)
                cols["volume"].append(result.volume)
                cols["date"].append(result.timestamp.astimezone(date_to.tzinfo))
                cols["exchange"].append("LIME")
                cols["exchange_country"].append("US")
                cols["symbol"].append(symbol)
        df = pl.DataFrame(cols, schema=[("open", pl.Float64()), ("close", pl.Float64()),
                                        ("price", pl.Float64()),
                                        ("high", pl.Float64()), ("low", pl.Float64()),
                                        ("volume", pl.Float64()),
                                        ("date", pl.Datetime), ("exchange", pl.String),
                                        ("exchange_country", pl.String), ("symbol", pl.String)
                                        ])
        return df.filter(pl.col("date") >= date_from, pl.col("date") <= date_to)

    async def get_data(self, symbols: list[str],
                       frequency: datetime.timedelta,
                       date_from: datetime.datetime,
                       date_to: datetime.datetime,
                       ) -> pl.DataFrame:

        quotes_tasks = [self._lime_sdk_client.market.get_quotes_history(
            symbol=symbol, period=self._frequency_to_period(frequency=frequency), from_date=date_from,
            to_date=date_to
        ) for symbol in symbols]

        results = await asyncio.gather(*quotes_tasks)

        for result in results:
            if len(df) > 0:
                df = df.rename(
                    {
                        "o": "open",
                        "h": "high",
                        "l": "low",
                        "c": "close",
                        "v": "volume",
                        "Date": "date"
                    }
                )
                df = df.with_columns(
                    pl.lit(symbol).alias("symbol"),
                    pl.lit("LIME").alias("exchange"),
                    pl.lit("US").alias("exchange_country"),
                    date=pl.col("date").dt.replace_time_zone(str(date_from.tzinfo)),
                ).filter(pl.col("date") >= date_from, pl.col("date") <= date_to)
                return df

    def get_data_by_limit(self, fields: frozenset[str],
                          limit: int,
                          end_date: datetime.datetime,
                          frequency: datetime.timedelta,
                          assets: frozenset[Asset],
                          include_end_date: bool,
                          exchange_name: str,
                          exchange_country: str,
                          trading_calendar: ExchangeCalendar
                          ) -> pl.DataFrame:

        symbols = [asset.get_symbol_by_exchange(exchange_name=exchange_name) for asset in assets if
                   asset.get_symbol_by_exchange(exchange_name=exchange_name)]
        # period = self._lime_trader_sdk_data_source._frequency_to_period(frequency=frequency)
        # lowest frequency of lime trader sdk is 1 minute
        try:
            sdk_period = self._frequency_to_period(frequency=frequency)
        except Exception as e:
            # TODO: handle case when we have for example 2d frequency, use 1 day with multiplier of 2
            # if datetime.timedelta(minutes=1) < frequency:
            #     multiplier = int(frequency / self.frequency)
            #     total_bar_count = limit * multiplier
            raise
        if frequency >= datetime.timedelta(days=1):
            multiplier = frequency // datetime.timedelta(days=1)
            time_window = trading_calendar.sessions_window(session=end_date.date(), count=-limit * multiplier)
            start_dt = time_window[0].replace(tzinfo=end_date.tzinfo)
            end_dt = time_window[-1].replace(tzinfo=end_date.tzinfo)

        else:
            multiplier = frequency // datetime.timedelta(minutes=1)
            time_window = trading_calendar.minutes_window(minute=end_date, count=-limit * multiplier)
            start_dt = time_window[0].astimezone(end_date.tzinfo)
            end_dt = time_window[-1].astimezone(end_date.tzinfo)

        sdk_results = [self._lime_sdk_client_sync.market.get_quotes_history(
            symbol=symbol, period=self._frequency_to_period(frequency=frequency), from_date=time_window[0],
            to_date=time_window[-1]
        ) for symbol in symbols]

        cols = {"open": [], "close": [], "price": [], "high": [], "low": [], "volume": [], "date": [], "exchange": [],
                "symbol": [], "exchange_country": []}
        for results, symbol in zip(sdk_results, symbols):
            for result in results:
                cols["open"].append(result.open)
                cols["close"].append(result.close)
                cols["price"].append(result.close)
                cols["high"].append(result.high)
                cols["low"].append(result.low)
                cols["volume"].append(result.volume)
                cols["date"].append(result.timestamp.astimezone(end_date.tzinfo))
                cols["exchange"].append(exchange_name)
                cols["exchange_country"].append(exchange_country)
                cols["symbol"].append(symbol)
        df_raw = pl.DataFrame(cols, schema=[("open", pl.Float64()), ("close", pl.Float64()),
                                            ("price", pl.Float64()),
                                            ("high", pl.Float64()), ("low", pl.Float64()),
                                            ("volume", pl.Float64()),
                                            ("date", pl.Datetime), ("exchange", pl.String),
                                            ("exchange_country", pl.String), ("symbol", pl.String)
                                            ])
        df_raw = df_raw.filter(pl.col("date") >= start_dt, pl.col("date") <= end_dt)
        if multiplier != 1:
            df = df_raw.group_by_dynamic(
                index_column="date", every=frequency, by="sid").agg(pl.col(field).last() for field in fields).tail(
                limit)
            return df.select(pl.col(col) for col in fields)

        return df_raw.select(pl.col(col) for col in fields)

    def _frequency_to_period(self, frequency: datetime.timedelta) -> Period:
        match frequency:
            case datetime.timedelta(seconds=60):
                return Period.MINUTE
            case datetime.timedelta(seconds=300):
                return Period.MINUTE_5
            case datetime.timedelta(seconds=900):
                return Period.MINUTE_15
            case datetime.timedelta(seconds=1800):
                return Period.MINUTE_30
            case datetime.timedelta(seconds=3600):
                return Period.HOUR
            case datetime.timedelta(days=1):
                return Period.DAY
            case datetime.timedelta(days=7):
                return Period.WEEK
            case datetime.timedelta(days=30):
                return Period.MONTH
            case datetime.timedelta(days=90):
                return Period.QUARTER
            case datetime.timedelta(days=365):
                return Period.YEAR
            case _:
                raise ValueError(f"Unknown frequency {frequency} for lime trader sdk")



from functools import total_ordering
from iso4217 import Currency as ISO4217Currency

_ALL_CURRENCIES = {}


@total_ordering
class Currency:
    """A currency identifier, as defined by ISO-4217.

    Parameters
    ----------
    code : str
        ISO-4217 code for the currency.

    Attributes
    ----------
    code : str
        ISO-4217 currency code for the currency, e.g., 'USD'.
    name : str
        Plain english name for the currency, e.g., 'US Dollar'.
    """

    def __new__(cls, code):
        try:
            return _ALL_CURRENCIES[code]
        except KeyError:
            if code is None:
                name = "NO CURRENCY"
            else:
                try:
                    name = ISO4217Currency(code).currency_name
                except ValueError as exc:
                    raise ValueError(
                        "{!r} is not a valid currency code.".format(code)
                    ) from exc

            obj = _ALL_CURRENCIES[code] = super(Currency, cls).__new__(cls)
            obj._code = code
            obj._name = name
            return obj

    @property
    def code(self):
        """ISO-4217 currency code for the currency.

        Returns
        -------
        code : str
        """
        return self._code

    @property
    def name(self):
        """Plain english name for the currency.

        Returns
        -------
        name : str
        """
        return self._name

    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        return self.code == other.code

    def __hash__(self):
        return hash(self.code)

    def __lt__(self, other):
        return self.code < other.code

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self.code)



import datetime
from typing import Any

import numpy as np

from ziplime.exchanges.exchange import Exchange
from ziplime.finance.domain.ledger import Ledger


class ReturnsStatistic:
    """A metric that reports an end of simulation scalar or time series
    computed from the algorithm returns.

    Parameters
    ----------
    function : callable
        The function to call on the daily returns.
    field_name : str, optional
        The name of the field. If not provided, it will be
        ``function.__name__``.
    """

    def __init__(self, function, field_name=None):
        if field_name is None:
            field_name = function.__name__

        self._function = function
        self._field_name = field_name

    def end_of_bar(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                   exchanges: dict[str, Exchange]):
        res = self._function(ledger.daily_returns_array[: session_ix + 1])
        if not np.isfinite(res):
            res = None
        packet["cumulative_risk_metrics"][self._field_name] = res

    end_of_session = end_of_bar



import datetime
from abc import abstractmethod

from sqlalchemy.orm import Mapped

from ziplime.core.db.annotated_types import AssetRouterFKPK, Decimal8, SymbolsUniverseFKPK
from ziplime.core.db.base_model import BaseModel


class SymbolsUniverseAssetModel(BaseModel):
    __tablename__ = "symbols_universe_assets"
    symbol_universe_id: Mapped[SymbolsUniverseFKPK]
    asset_sid: Mapped[AssetRouterFKPK]
    # asset_router = relationship(AssetRouter)
    ratio: Mapped[Decimal8 | None]



import datetime
from typing import Any

from ziplime.exchanges.exchange import Exchange
from ziplime.finance.domain.ledger import Ledger


class Positions:
    """Tracks daily positions."""

    def end_of_bar(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                   exchanges: dict[str, Exchange]):
        packet["minute_perf"]["positions"] = ledger.positions(session)

    def end_of_session(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                       exchanges: dict[str, Exchange]):
        packet["daily_perf"]["positions"] = ledger.positions(dt=session)



import datetime
from dataclasses import dataclass, field

from ziplime.assets.entities.asset import Asset
from ziplime.domain.position import Position


@dataclass
class Portfolio:
    # capital_used: float
    cash_flow: float
    starting_cash: float
    portfolio_value: float
    pnl: float
    returns: float
    cash: float
    positions_value: float
    positions_exposure: float
    positions: dict[Asset, Position] = field(default_factory=dict)
    start_date: datetime.datetime | None = None





import empyrical

from ziplime.finance.metrics.max_leverage import MaxLeverage
from ziplime.finance.metrics.alpha_beta import AlphaBeta
from ziplime.finance.metrics.benchmark_returns_and_volatility import BenchmarkReturnsAndVolatility
from ziplime.finance.metrics.cash_flow import CashFlow
from ziplime.finance.metrics.constant_cumulative_risk_metric import ConstantCumulativeRiskMetric
from ziplime.finance.metrics.daily_ledger_field import DailyLedgerField
from ziplime.finance.metrics.num_trading_days import NumTradingDays
from ziplime.finance.metrics.orders import Orders
from ziplime.finance.metrics.period_label import PeriodLabel
from ziplime.finance.metrics.pnl import PNL
from ziplime.finance.metrics.retruns_statistics import ReturnsStatistic
from ziplime.finance.metrics.simple_ledger_field import SimpleLedgerField
from ziplime.finance.metrics.start_of_period_ledger_field import StartOfPeriodLedgerField
from ziplime.finance.metrics.transactions import Transactions
from ziplime.finance.metrics.returns import Returns


def default_metrics():
    return {
        Returns(),
        ReturnsStatistic(empyrical.annual_volatility, "algo_volatility"),
        BenchmarkReturnsAndVolatility(),
        PNL(),
        CashFlow(),
        Orders(),
        Transactions(),
        SimpleLedgerField("positions"),
        StartOfPeriodLedgerField(
            "portfolio.positions_exposure",
            "starting_exposure",
        ),
        DailyLedgerField(
            "portfolio.positions_exposure",
            "ending_exposure",
        ),
        StartOfPeriodLedgerField("portfolio.positions_value", "starting_value"),
        DailyLedgerField("portfolio.positions_value", "ending_value"),
        StartOfPeriodLedgerField("portfolio.cash", "starting_cash"),
        DailyLedgerField("portfolio.cash", "ending_cash"),
        DailyLedgerField("portfolio.portfolio_value"),
        DailyLedgerField("position_tracker.stats.longs_count"),
        DailyLedgerField("position_tracker.stats.shorts_count"),
        DailyLedgerField("position_tracker.stats.long_value"),
        DailyLedgerField("position_tracker.stats.short_value"),
        DailyLedgerField("position_tracker.stats.long_exposure"),
        DailyLedgerField("position_tracker.stats.short_exposure"),
        DailyLedgerField("account.gross_leverage"),
        DailyLedgerField("account.net_leverage"),
        AlphaBeta(),
        ReturnsStatistic(empyrical.sharpe_ratio, "sharpe"),
        ReturnsStatistic(empyrical.sortino_ratio, "sortino"),
        ReturnsStatistic(empyrical.max_drawdown),
        MaxLeverage(),
        # Please kill these!
        ConstantCumulativeRiskMetric("excess_return", 0.0),
        ConstantCumulativeRiskMetric("treasury_period_return", 0.0),
        NumTradingDays(),
        PeriodLabel(),
    }



from sqlalchemy.orm import Mapped

from ziplime.core.db.annotated_types import DateIndexed, IntegerIndexed, StringPK
from ziplime.core.db.base_model import BaseModel


class Split(BaseModel):
    __tablename__ = "splits"

    index: Mapped[StringPK]
    sid: Mapped[IntegerIndexed]

    effective_date: Mapped[DateIndexed]
    ratio: Mapped[float]



from typing import Callable

from exchange_calendars import ExchangeCalendar

from ziplime.assets.domain.asset_type import AssetType
from ziplime.assets.entities.asset import Asset
from ziplime.assets.entities.futures_contract import FuturesContract
from ziplime.finance.asset_restrictions import Restrictions
from ziplime.finance.domain.order import Order
from ziplime.finance.execution import ExecutionStyle
from ziplime.utils.events import EventRule


class BaseTradingAlgorithm:
    def schedule_function(
            self,
            func: Callable,
            date_rule: EventRule = None,
            time_rule: EventRule = None,
            half_days: bool = True,
            calendar: ExchangeCalendar | None = None,
    ): ...

    def record(self, *args, **kwargs): ...

    def continuous_future(
            self, root_symbol_str: str, offset: int = 0, roll: str = "volume", adjustment: str = "mul"
    ): ...

    async def symbol(
            self, symbol_str: str, asset_type: AssetType = AssetType.EQUITY,
            exchange_name: str = None,
            country_code: str | None = None
    ) -> Asset | None: ...

    def symbols(self, *args, **kwargs): ...

    async def sid(self, sid: int) -> Asset | None: ...

    async def future_symbol(self, symbol: str, exchange_name: str = None) -> FuturesContract | None: ...

    async def order(self, asset: Asset, amount: int, style: ExecutionStyle,
                    exchange_name: str | None = None) -> Order | None: ...

    async def order_value(self, asset: Asset, value: float, limit_price: float | None = None,
                          stop_price: float | None = None,
                          style: ExecutionStyle | None = None): ...

    def get_datetime(self): ...

    def set_slippage(self, us_equities=None, us_futures=None): ...

    def set_commission(self, us_equities=None, us_futures=None): ...

    def set_cancel_policy(self, cancel_policy): ...

    def set_symbol_lookup_date(self, dt): ...

    async def order_percent(
            self, asset: Asset, percent: float, style: ExecutionStyle
    ): ...

    async def order_target(
            self, asset: Asset, target: int, style: ExecutionStyle
    ): ...

    async def order_target_value(
            self, asset: Asset, target: float, style: ExecutionStyle
    ): ...

    async def order_target_percent(
            self, asset: Asset, target: float,
            style: ExecutionStyle
    ): ...

    def get_open_orders(self, asset=None): ...

    def get_order(self, order_id) -> Order | None: ...

    def cancel_order(self, order_id: str, relay_status: bool = True) -> None: ...

    def set_max_leverage(self, max_leverage): ...

    def set_min_leverage(self, min_leverage, grace_period): ...

    def set_max_position_size(
            self, asset=None, max_shares=None, max_notional=None, on_error="fail"
    ): ...

    def set_max_order_size(
            self, asset=None, max_shares=None, max_notional=None, on_error="fail"
    ): ...

    def set_max_order_count(self, max_count, on_error="fail"): ...

    def set_asset_restrictions(self, restrictions: Restrictions, on_error: str = "fail"): ...

    def set_long_only(self, on_error="fail"):...

    def attach_pipeline(self, pipeline, name, chunks=None, eager=True): ...

    def pipeline_output(self, name): ...



from .equity_pricing import EquityPricing, USEquityPricing
from .dataset import (
    BoundColumn,
    Column,
    DataSet,
    DataSetFamily,
    DataSetFamilySlice,
)

__all__ = [
    "BoundColumn",
    "Column",
    "DataSet",
    "EquityPricing",
    "DataSetFamily",
    "DataSetFamilySlice",
    "USEquityPricing",
]



"""Synthetic data loaders for testing."""

import numpy as np

from numpy.random import RandomState
from pandas import DataFrame, Timestamp
from sqlite3 import connect as sqlite3_connect

from .base import PipelineLoader
from .frame import DataFrameLoader


from ziplime.utils.numpy_utils import (
    bool_dtype,
    datetime64ns_dtype,
    float64_dtype,
    int64_dtype,
    object_dtype,
)
from ...assets.repositories.sqlalchemy_adjustments_repository import SqlAlchemyAdjustmentRepository

US_EQUITY_PRICING_BCOLZ_COLUMNS = (
    "open",
    "high",
    "low",
    "close",
    "volume",
    "day",
    "id",
)

UINT_32_MAX = np.iinfo(np.uint32).max


def nanos_to_seconds(nanos):
    return nanos / (1000 * 1000 * 1000)


class PrecomputedLoader(PipelineLoader):
    """Synthetic PipelineLoader that uses a pre-computed array for each column.

    Parameters
    ----------
    values : dict
        Map from column to values to use for that column.
        Values can be anything that can be passed as the first positional
        argument to a DataFrame whose indices are ``dates`` and ``sids``
    dates : iterable[datetime-like]
        Row labels for input data.  Can be anything that pd.DataFrame will
        coerce to a DatetimeIndex.
    sids : iterable[int-like]
        Column labels for input data.  Can be anything that pd.DataFrame will
        coerce to an Int64Index.

    Notes
    -----
    Adjustments are unsupported by this loader.
    """

    def __init__(self, constants, dates, sids):
        loaders = {}
        for column, const in constants.items():
            frame = DataFrame(
                const,
                index=dates,
                columns=sids,
                dtype=column.dtype,
            )
            loaders[column] = DataFrameLoader(
                column=column,
                baseline=frame,
                adjustments=None,
            )

        self._loaders = loaders

    def load_adjusted_array(self, domain, columns, dates, sids, mask):
        """Load by delegating to sub-loaders."""
        out = {}
        for col in columns:
            try:
                loader = self._loaders.get(col)
                if loader is None:
                    loader = self._loaders[col.unspecialize()]
            except KeyError as exc:
                raise ValueError("Couldn't find loader for %s" % col) from exc
            out.update(loader.load_adjusted_array(domain, [col], dates, sids, mask))
        return out


class EyeLoader(PrecomputedLoader):
    """A PrecomputedLoader that emits arrays containing 1s on the diagonal and 0s
    elsewhere.

    Parameters
    ----------
    columns : list[BoundColumn]
        Columns that this loader should know about.
    dates : iterable[datetime-like]
        Same as PrecomputedLoader.
    sids : iterable[int-like]
        Same as PrecomputedLoader
    """

    def __init__(self, columns, dates, sids):
        shape = (len(dates), len(sids))
        super(EyeLoader, self).__init__(
            {column: np.eye(shape, dtype=column.dtype) for column in columns},
            dates,
            sids,
        )


class SeededRandomLoader(PrecomputedLoader):
    """A PrecomputedLoader that emits arrays randomly-generated with a given seed.

    Parameters
    ----------
    seed : int
        Seed for numpy.random.RandomState.
    columns : list[BoundColumn]
        Columns that this loader should know about.
    dates : iterable[datetime-like]
        Same as PrecomputedLoader.
    sids : iterable[int-like]
        Same as PrecomputedLoader
    """

    def __init__(self, seed, columns, dates, sids):
        self._seed = seed
        super(SeededRandomLoader, self).__init__(
            {c: self.values(c.dtype, dates, sids) for c in columns},
            dates,
            sids,
        )

    def values(self, dtype, dates, sids):
        """Make a random array of shape (len(dates), len(sids)) with ``dtype``."""
        shape = (len(dates), len(sids))
        return {
            datetime64ns_dtype: self._datetime_values,
            float64_dtype: self._float_values,
            int64_dtype: self._int_values,
            bool_dtype: self._bool_values,
            object_dtype: self._object_values,
        }[dtype](shape)

    @property
    def state(self):
        """Make a new RandomState from our seed.

        This ensures that every call to _*_values produces the same output
        every time for a given SeededRandomLoader instance.
        """
        return RandomState(self._seed)

    def _float_values(self, shape):
        """Return uniformly-distributed floats between -0.0 and 100.0."""
        return self.state.uniform(low=0.0, high=100.0, size=shape)

    def _int_values(self, shape):
        """
        Return uniformly-distributed integers between 0 and 100.
        """
        return self.state.randint(low=0, high=100, size=shape).astype(
            "int64"
        )  # default is system int

    def _datetime_values(self, shape):
        """Return uniformly-distributed dates in 2014."""
        start = Timestamp("2014", tz="UTC").asm8
        offsets = self.state.randint(
            low=0,
            high=364,
            size=shape,
        ).astype("timedelta64[D]")
        return start + offsets

    def _bool_values(self, shape):
        """Return uniformly-distributed True/False values."""
        return self.state.randn(*shape) < 0

    def _object_values(self, shape):
        res = self._int_values(shape).astype(str).astype(object)
        return res


OHLCV = ("open", "high", "low", "close", "volume")
OHLC = ("open", "high", "low", "close")

PSEUDO_EPOCH_UTC = Timestamp("2000-01-01", tz="UTC")
PSEUDO_EPOCH_NAIVE = Timestamp("2000-01-01")

# TODO FIX TZ MESS


def asset_start(asset_info, asset, tz=None):
    ret = asset_info.loc[asset]["start_date"]
    if tz is not None:
        ret = ret.tz_localize(tz)
    # assert ret.tzname() == "UTC", "Unexpected non-UTC timestamp"
    return ret


def asset_end(asset_info, asset, tz=None):
    ret = asset_info.loc[asset]["end_date"]
    if tz is not None:
        ret = ret.tz_localize(tz)
    # if ret.tz is None:
    #     ret = ret.tz_localize("UTC")
    # assert ret.tzname() == "UTC", "Unexpected non-UTC timestamp"
    return ret


def make_bar_data(asset_info, calendar, holes=None):
    """For a given asset/date/column combination, we generate a corresponding raw
    value using the following formula for OHLCV columns:

    data(asset, date, column) = (100,000 * asset_id)
                              + (10,000 * column_num)
                              + (date - Jan 1 2000).days  # ~6000 for 2015
    where:
        column_num('open') = 0
        column_num('high') = 1
        column_num('low') = 2
        column_num('close') = 3
        column_num('volume') = 4

    We use days since Jan 1, 2000 to guarantee that there are no collisions
    while also the produced values smaller than UINT32_MAX / 1000.

    For 'day' and 'id', we use the standard format expected by the base class.

    Parameters
    ----------
    asset_info : DataFrame
        DataFrame with asset_id as index and 'start_date'/'end_date' columns.
    calendar : pd.DatetimeIndex
        The trading calendar to use.
    holes : dict[int -> tuple[pd.Timestamps]], optional
        A dict mapping asset ids to the tuple of dates that should have
        no data for that asset in the output. Default is no holes.

    Yields
    ------
    p : (int, pd.DataFrame)
        A sid, data pair to be passed to BcolzDailyDailyBarWriter.write
    """
    assert (
        # Using .value here to avoid having to care about UTC-aware dates.
        PSEUDO_EPOCH_UTC.value
        < calendar.normalize().min().value
        <= asset_info["start_date"].min().value
    ), "calendar.min(): %s\nasset_info['start_date'].min(): %s" % (
        calendar.min(),
        asset_info["start_date"].min(),
    )

    assert (asset_info["start_date"] < asset_info["end_date"]).all()

    def _raw_data_for_asset(asset_id):
        """Generate 'raw' data that encodes information about the asset.

        See docstring for a description of the data format.
        """
        # Get the dates for which this asset existed according to our asset
        # info.
        datetimes = calendar[
            calendar.slice_indexer(
                asset_start(asset_info, asset_id, tz=calendar.tz),
                asset_end(asset_info, asset_id, tz=calendar.tz),
            )
        ]

        data = np.full(
            (len(datetimes), len(US_EQUITY_PRICING_BCOLZ_COLUMNS)),
            asset_id * 100 * 1000,
            dtype=np.uint32,
        )

        # Add 10,000 * column-index to OHLCV columns
        data[:, :5] += np.arange(5, dtype=np.uint32) * 1000

        # Add days since Jan 1 2001 for OHLCV columns.
        # TODO FIXME TZ MESS

        if datetimes.tzinfo is None:
            data[:, :5] += np.array(
                (datetimes.tz_localize("UTC") - PSEUDO_EPOCH_UTC).days
            )[:, None].astype(np.uint32)
        else:
            data[:, :5] += np.array((datetimes - PSEUDO_EPOCH_UTC).days)[
                :, None
            ].astype(np.uint32)

        frame = DataFrame(
            data,
            index=datetimes,
            columns=US_EQUITY_PRICING_BCOLZ_COLUMNS,
        )

        if holes is not None and asset_id in holes:
            for dt in holes[asset_id]:
                frame.loc[dt, OHLC] = np.nan
                frame.loc[dt, ["volume"]] = 0

        frame["day"] = nanos_to_seconds(datetimes.asi8)
        frame["id"] = asset_id
        return frame

    for asset in asset_info.index:
        yield asset, _raw_data_for_asset(asset)


def expected_bar_value(asset_id, date, colname):
    """Check that the raw value for an asset/date/column triple is as
    expected.

    Used by tests to verify data written by a writer.
    """
    from_asset = asset_id * 100000
    from_colname = OHLCV.index(colname) * 1000
    from_date = (date - PSEUDO_EPOCH_NAIVE.tz_localize(date.tzinfo)).days
    return from_asset + from_colname + from_date


def expected_bar_value_with_holes(asset_id, date, colname, holes, missing_value):
    # Explicit holes are filled with the missing value.
    if asset_id in holes and date in holes[asset_id]:
        return missing_value

    return expected_bar_value(asset_id, date, colname)


def expected_bar_values_2d(dates, assets, asset_info, colname, holes=None):
    """Return an 2D array containing cls.expected_value(asset_id, date,
    colname) for each date/asset pair in the inputs.

    Missing locs are filled with 0 for volume and NaN for price columns:

        - Values before/after an asset's lifetime.
        - Values for asset_ids not contained in asset_info.
        - Locs defined in `holes`.
    """
    if colname == "volume":
        dtype = np.uint32
        missing = 0
    else:
        dtype = np.float64
        missing = float("nan")

    data = np.full((len(dates), len(assets)), missing, dtype=dtype)
    for j, asset in enumerate(assets):
        # Use missing values when asset_id is not contained in asset_info.
        if asset not in asset_info.index:
            continue

        start = asset_start(asset_info, asset)
        end = asset_end(asset_info, asset)
        for i, date in enumerate(dates):
            # No value expected for dates outside the asset's start/end
            # date.
            # TODO FIXME TZ MESS
            if not (
                start.tz_localize(date.tzinfo) <= date <= end.tz_localize(date.tzinfo)
            ):
                continue

            if holes is not None:
                expected = expected_bar_value_with_holes(
                    asset,
                    date,
                    colname,
                    holes,
                    missing,
                )
            else:
                expected = expected_bar_value(asset, date, colname)

            data[i, j] = expected
    return data


class NullAdjustmentReader(SqlAlchemyAdjustmentRepository):
    """A SQLiteAdjustmentReader that stores no adjustments and uses in-memory
    SQLite.
    """

    def __init__(self):
        conn = sqlite3_connect(":memory:")
        writer = SqlAlchemyAdjustmentRepository(conn, None, None)
        empty = DataFrame(
            {
                "sid": np.array([], dtype=np.uint32),
                "effective_date": np.array([], dtype=np.uint32),
                "ratio": np.array([], dtype=float),
            }
        )
        empty_dividends = DataFrame(
            {
                "sid": np.array([], dtype=np.uint32),
                "amount": np.array([], dtype=np.float64),
                "record_date": np.array([], dtype="datetime64[ns]"),
                "ex_date": np.array([], dtype="datetime64[ns]"),
                "declared_date": np.array([], dtype="datetime64[ns]"),
                "pay_date": np.array([], dtype="datetime64[ns]"),
            }
        )
        writer.write(splits=empty, mergers=empty, dividends=empty_dividends)
        super(NullAdjustmentReader, self).__init__(conn)



from dataclasses import dataclass

from sqlalchemy.orm import Mapped

from ziplime.assets.entities.asset import Asset


@dataclass(frozen=True)
class SymbolsUniverse:
    assets: list[Asset]
    name: Mapped[str]
    symbol: Mapped[str]
    universe_type: Mapped[str]



from abc import abstractmethod

import numpy as np
import pandas as pd
from toolz import groupby

from ziplime.lib.adjusted_array import AdjustedArray
from ziplime.lib.adjustment import (
    Datetime641DArrayOverwrite,
    Datetime64Overwrite,
    Float641DArrayOverwrite,
    Float64Multiply,
    Float64Overwrite,
)

from ziplime.pipeline.common import (
    EVENT_DATE_FIELD_NAME,
    FISCAL_QUARTER_FIELD_NAME,
    FISCAL_YEAR_FIELD_NAME,
    SID_FIELD_NAME,
    TS_FIELD_NAME,
)
from ziplime.pipeline.loaders.base import PipelineLoader
from ziplime.utils.date_utils import make_utc_aware
from ziplime.utils.numpy_utils import datetime64ns_dtype, float64_dtype
from ziplime.pipeline.loaders.utils import (
    ffill_across_cols,
    last_in_date_group,
)

INVALID_NUM_QTRS_MESSAGE = (
    "Passed invalid number of quarters %s; " "must pass a number of quarters >= 0"
)
NEXT_FISCAL_QUARTER = "next_fiscal_quarter"
NEXT_FISCAL_YEAR = "next_fiscal_year"
NORMALIZED_QUARTERS = "normalized_quarters"
PREVIOUS_FISCAL_QUARTER = "previous_fiscal_quarter"
PREVIOUS_FISCAL_YEAR = "previous_fiscal_year"
SHIFTED_NORMALIZED_QTRS = "shifted_normalized_quarters"
SIMULATION_DATES = "dates"


def normalize_quarters(years, quarters):
    return years * 4 + quarters - 1


def split_normalized_quarters(normalized_quarters):
    years = normalized_quarters // 4
    quarters = normalized_quarters % 4
    return years, quarters + 1


# These metadata columns are used to align event indexers.
metadata_columns = frozenset(
    {
        TS_FIELD_NAME,
        SID_FIELD_NAME,
        EVENT_DATE_FIELD_NAME,
        FISCAL_QUARTER_FIELD_NAME,
        FISCAL_YEAR_FIELD_NAME,
    }
)


def required_estimates_fields(columns):
    """Compute the set of resource columns required to serve
    `columns`.
    """
    # We also expect any of the field names that our loadable columns
    # are mapped to.
    return metadata_columns.union(columns.values())


def validate_column_specs(events, columns):
    """Verify that the columns of ``events`` can be used by a
    EarningsEstimatesLoader to serve the BoundColumns described by
    `columns`.
    """
    required = required_estimates_fields(columns)
    received = set(events.columns)
    missing = required - received
    if missing:
        raise ValueError(
            "EarningsEstimatesLoader missing required columns {missing}.\n"
            "Got Columns: {received}\n"
            "Expected Columns: {required}".format(
                missing=sorted(missing),
                received=sorted(received),
                required=sorted(required),
            )
        )


def add_new_adjustments(adjustments_dict, adjustments, column_name, ts):
    try:
        adjustments_dict[column_name][ts].extend(adjustments)
    except KeyError:
        adjustments_dict[column_name][ts] = adjustments


class EarningsEstimatesLoader(PipelineLoader):
    """An abstract pipeline loader for estimates data that can load data a
    variable number of quarters forwards/backwards from calendar dates
    depending on the `num_announcements` attribute of the columns' dataset.
    If split adjustments are to be applied, a loader, split-adjusted columns,
    and the split-adjusted asof-date must be supplied.

    Parameters
    ----------
    estimates : pd.DataFrame
        The raw estimates data; must contain at least 5 columns:
            sid : int64
                The asset id associated with each estimate.

            event_date : datetime64[ns]
                The date on which the event that the estimate is for will/has
                occurred.

            timestamp : datetime64[ns]
                The datetime where we learned about the estimate.

            fiscal_quarter : int64
                The quarter during which the event has/will occur.

            fiscal_year : int64
                The year during which the event has/will occur.

    name_map : dict[str -> str]
        A map of names of BoundColumns that this loader will load to the
        names of the corresponding columns in `events`.
    """

    def __init__(self, estimates, name_map):
        validate_column_specs(estimates, name_map)

        self.estimates = estimates[
            estimates[EVENT_DATE_FIELD_NAME].notnull()
            & estimates[FISCAL_QUARTER_FIELD_NAME].notnull()
            & estimates[FISCAL_YEAR_FIELD_NAME].notnull()
        ]
        self.estimates[NORMALIZED_QUARTERS] = normalize_quarters(
            self.estimates[FISCAL_YEAR_FIELD_NAME],
            self.estimates[FISCAL_QUARTER_FIELD_NAME],
        )

        self.array_overwrites_dict = {
            datetime64ns_dtype: Datetime641DArrayOverwrite,
            float64_dtype: Float641DArrayOverwrite,
        }
        self.scalar_overwrites_dict = {
            datetime64ns_dtype: Datetime64Overwrite,
            float64_dtype: Float64Overwrite,
        }

        self.name_map = name_map

    @abstractmethod
    def get_zeroth_quarter_idx(self, stacked_last_per_qtr):
        raise NotImplementedError("get_zeroth_quarter_idx")

    @abstractmethod
    def get_shifted_qtrs(self, zero_qtrs, num_announcements):
        raise NotImplementedError("get_shifted_qtrs")

    @abstractmethod
    def create_overwrite_for_estimate(
        self,
        column,
        column_name,
        last_per_qtr,
        next_qtr_start_idx,
        requested_quarter,
        sid,
        sid_idx,
        col_to_split_adjustments,
        split_adjusted_asof_idx,
    ):
        raise NotImplementedError("create_overwrite_for_estimate")

    @property
    @abstractmethod
    def searchsorted_side(self):
        return NotImplementedError("searchsorted_side")

    def get_requested_quarter_data(
        self,
        zero_qtr_data,
        zeroth_quarter_idx,
        stacked_last_per_qtr,
        num_announcements,
        dates,
    ):
        """Selects the requested data for each date.

        Parameters
        ----------
        zero_qtr_data : pd.DataFrame
            The 'time zero' data for each calendar date per sid.
        zeroth_quarter_idx : pd.Index
            An index of calendar dates, sid, and normalized quarters, for only
            the rows that have a next or previous earnings estimate.
        stacked_last_per_qtr : pd.DataFrame
            The latest estimate known with the dates, normalized quarter, and
            sid as the index.
        num_announcements : int
            The number of annoucements out the user requested relative to
            each date in the calendar dates.
        dates : pd.DatetimeIndex
            The calendar dates for which estimates data is requested.

        Returns
        --------
        requested_qtr_data : pd.DataFrame
            The DataFrame with the latest values for the requested quarter
            for all columns; `dates` are the index and columns are a MultiIndex
            with sids at the top level and the dataset columns on the bottom.
        """
        zero_qtr_data_idx = zero_qtr_data.index
        requested_qtr_idx = pd.MultiIndex.from_arrays(
            [
                zero_qtr_data_idx.get_level_values(0),
                zero_qtr_data_idx.get_level_values(1),
                self.get_shifted_qtrs(
                    zeroth_quarter_idx.get_level_values(
                        NORMALIZED_QUARTERS,
                    ),
                    num_announcements,
                ),
            ],
            names=[
                zero_qtr_data_idx.names[0],
                zero_qtr_data_idx.names[1],
                SHIFTED_NORMALIZED_QTRS,
            ],
        )

        requested_qtr_data = stacked_last_per_qtr.reindex(index=requested_qtr_idx)
        requested_qtr_data = requested_qtr_data.reset_index(
            SHIFTED_NORMALIZED_QTRS,
        )
        # Calculate the actual year/quarter being requested and add those in
        # as columns.
        (
            requested_qtr_data[FISCAL_YEAR_FIELD_NAME],
            requested_qtr_data[FISCAL_QUARTER_FIELD_NAME],
        ) = split_normalized_quarters(requested_qtr_data[SHIFTED_NORMALIZED_QTRS])
        # Once we're left with just dates as the index, we can reindex by all
        # dates so that we have a value for each calendar date.
        return requested_qtr_data.unstack(SID_FIELD_NAME).reindex(dates)

    def get_split_adjusted_asof_idx(self, dates):
        """Compute the index in `dates` where the split-adjusted-asof-date
        falls. This is the date up to which, and including which, we will
        need to unapply all adjustments for and then re-apply them as they
        come in. After this date, adjustments are applied as normal.

        Parameters
        ----------
        dates : pd.DatetimeIndex
            The calendar dates over which the Pipeline is being computed.

        Returns
        -------
        split_adjusted_asof_idx : int
            The index in `dates` at which the data should be split.
        """
        split_adjusted_asof_idx = dates.searchsorted(self._split_adjusted_asof)
        # make_utc_aware(pd.DatetimeIndex(self._split_adjusted_asof))
        # The split-asof date is after the date index.
        if split_adjusted_asof_idx == len(dates):
            split_adjusted_asof_idx = len(dates) - 1
        if self._split_adjusted_asof.tzinfo is not None:
            if self._split_adjusted_asof < dates[0]:
                split_adjusted_asof_idx = -1
        else:
            if self._split_adjusted_asof < dates[0]:
                split_adjusted_asof_idx = -1
        return split_adjusted_asof_idx

    def collect_overwrites_for_sid(
        self,
        group,
        dates,
        requested_qtr_data,
        last_per_qtr,
        sid_idx,
        columns,
        all_adjustments_for_sid,
        sid,
    ):
        """Given a sid, collect all overwrites that should be applied for this
        sid at each quarter boundary.

        Parameters
        ----------
        group : pd.DataFrame
            The data for `sid`.
        dates : pd.DatetimeIndex
            The calendar dates for which estimates data is requested.
        requested_qtr_data : pd.DataFrame
            The DataFrame with the latest values for the requested quarter
            for all columns.
        last_per_qtr : pd.DataFrame
            A DataFrame with a column MultiIndex of [self.estimates.columns,
            normalized_quarters, sid] that allows easily getting the timeline
            of estimates for a particular sid for a particular quarter.
        sid_idx : int
            The sid's index in the asset index.
        columns : list of BoundColumn
            The columns for which the overwrites should be computed.
        all_adjustments_for_sid : dict[int -> AdjustedArray]
            A dictionary of the integer index of each timestamp into the date
            index, mapped to adjustments that should be applied at that
            index for the given sid (`sid`). This dictionary is modified as
            adjustments are collected.
        sid : int
            The sid for which overwrites should be computed.
        """
        # If data was requested for only 1 date, there can never be any
        # overwrites, so skip the extra work.
        if len(dates) == 1:
            return

        next_qtr_start_indices = dates.searchsorted(
            pd.DatetimeIndex(group[EVENT_DATE_FIELD_NAME]),
            side=self.searchsorted_side,
        )

        qtrs_with_estimates = group.index.get_level_values(NORMALIZED_QUARTERS).values
        for idx in next_qtr_start_indices:
            if 0 < idx < len(dates):
                # Find the quarter being requested in the quarter we're
                # crossing into.
                requested_quarter = requested_qtr_data[
                    SHIFTED_NORMALIZED_QTRS,
                    sid,
                ].iloc[idx]
                # Only add adjustments if the next quarter starts somewhere
                # in our date index for this sid. Our 'next' quarter can
                # never start at index 0; a starting index of 0 means that
                # the next quarter's event date was NaT.
                self.create_overwrites_for_quarter(
                    all_adjustments_for_sid,
                    idx,
                    last_per_qtr,
                    qtrs_with_estimates,
                    requested_quarter,
                    sid,
                    sid_idx,
                    columns,
                )

    def get_adjustments_for_sid(
        self,
        group,
        dates,
        requested_qtr_data,
        last_per_qtr,
        sid_to_idx,
        columns,
        col_to_all_adjustments,
        **kwargs,
    ):
        """

        Parameters
        ----------
        group : pd.DataFrame
            The data for the given sid.
        dates : pd.DatetimeIndex
            The calendar dates for which estimates data is requested.
        requested_qtr_data : pd.DataFrame
            The DataFrame with the latest values for the requested quarter
            for all columns.
        last_per_qtr : pd.DataFrame
            A DataFrame with a column MultiIndex of [self.estimates.columns,
            normalized_quarters, sid] that allows easily getting the timeline
            of estimates for a particular sid for a particular quarter.
        sid_to_idx : dict[int -> int]
            A dictionary mapping sid to he sid's index in the asset index.
        columns : list of BoundColumn
            The columns for which the overwrites should be computed.
        col_to_all_adjustments : dict[int -> AdjustedArray]
            A dictionary of the integer index of each timestamp into the date
            index, mapped to adjustments that should be applied at that
            index. This dictionary is for adjustments for ALL sids. It is
            modified as adjustments are collected.
        kwargs :
            Additional arguments used in collecting adjustments; unused here.
        """
        # Collect all adjustments for a given sid.
        all_adjustments_for_sid = {}
        sid = int(group.name)
        self.collect_overwrites_for_sid(
            group,
            dates,
            requested_qtr_data,
            last_per_qtr,
            sid_to_idx[sid],
            columns,
            all_adjustments_for_sid,
            sid,
        )
        self.merge_into_adjustments_for_all_sids(
            all_adjustments_for_sid, col_to_all_adjustments
        )

    def merge_into_adjustments_for_all_sids(
        self, all_adjustments_for_sid, col_to_all_adjustments
    ):
        """Merge adjustments for a particular sid into a dictionary containing
        adjustments for all sids.

        Parameters
        ----------
        all_adjustments_for_sid : dict[int -> AdjustedArray]
            All adjustments for a particular sid.
        col_to_all_adjustments : dict[int -> AdjustedArray]
            All adjustments for all sids.
        """

        for col_name in all_adjustments_for_sid:
            if col_name not in col_to_all_adjustments:
                col_to_all_adjustments[col_name] = {}
            for ts in all_adjustments_for_sid[col_name]:
                adjs = all_adjustments_for_sid[col_name][ts]
                add_new_adjustments(col_to_all_adjustments, adjs, col_name, ts)

    def get_adjustments(
        self,
        zero_qtr_data,
        requested_qtr_data,
        last_per_qtr,
        dates,
        assets,
        columns,
        **kwargs,
    ):
        """Creates an AdjustedArray from the given estimates data for the given
        dates.

        Parameters
        ----------
        zero_qtr_data : pd.DataFrame
            The 'time zero' data for each calendar date per sid.
        requested_qtr_data : pd.DataFrame
            The requested quarter data for each calendar date per sid.
        last_per_qtr : pd.DataFrame
            A DataFrame with a column MultiIndex of [self.estimates.columns,
            normalized_quarters, sid] that allows easily getting the timeline
            of estimates for a particular sid for a particular quarter.
        dates : pd.DatetimeIndex
            The calendar dates for which estimates data is requested.
        assets : pd.Int64Index
            An index of all the assets from the raw data.
        columns : list of BoundColumn
            The columns for which adjustments need to be calculated.
        kwargs :
            Additional keyword arguments that should be forwarded to
            `get_adjustments_for_sid` and to be used in computing adjustments
            for each sid.

        Returns
        -------
        col_to_all_adjustments : dict[int -> AdjustedArray]
            A dictionary of all adjustments that should be applied.
        """

        zero_qtr_data.sort_index(inplace=True)
        # Here we want to get the LAST record from each group of records
        # corresponding to a single quarter. This is to ensure that we select
        # the most up-to-date event date in case the event date changes.
        quarter_shifts = zero_qtr_data.groupby(
            level=[SID_FIELD_NAME, NORMALIZED_QUARTERS]
        ).nth(-1)

        col_to_all_adjustments = {}
        sid_to_idx = dict(zip(assets, range(len(assets))))
        quarter_shifts.groupby(level=SID_FIELD_NAME).apply(
            self.get_adjustments_for_sid,
            dates,
            requested_qtr_data,
            last_per_qtr,
            sid_to_idx,
            columns,
            col_to_all_adjustments,
            **kwargs,
        )
        return col_to_all_adjustments

    def create_overwrites_for_quarter(
        self,
        col_to_overwrites,
        next_qtr_start_idx,
        last_per_qtr,
        quarters_with_estimates_for_sid,
        requested_quarter,
        sid,
        sid_idx,
        columns,
    ):
        """Add entries to the dictionary of columns to adjustments for the given
        sid and the given quarter.

        Parameters
        ----------
        col_to_overwrites : dict [column_name -> list of ArrayAdjustment]
            A dictionary mapping column names to all overwrites for those
            columns.
        next_qtr_start_idx : int
            The index of the first day of the next quarter in the calendar
            dates.
        last_per_qtr : pd.DataFrame
            A DataFrame with a column MultiIndex of [self.estimates.columns,
            normalized_quarters, sid] that allows easily getting the timeline
            of estimates for a particular sid for a particular quarter; this
            is particularly useful for getting adjustments for 'next'
            estimates.
        quarters_with_estimates_for_sid : np.array
            An array of all quarters for which there are estimates for the
            given sid.
        requested_quarter : float
            The quarter for which the overwrite should be created.
        sid : int
            The sid for which to create overwrites.
        sid_idx : int
            The index of the sid in `assets`.
        columns : list of BoundColumn
            The columns for which to create overwrites.
        """
        for col in columns:
            column_name = self.name_map[col.name]
            if column_name not in col_to_overwrites:
                col_to_overwrites[column_name] = {}
            # If there are estimates for the requested quarter,
            # overwrite all values going up to the starting index of
            # that quarter with estimates for that quarter.
            if requested_quarter in quarters_with_estimates_for_sid:
                adjs = self.create_overwrite_for_estimate(
                    col,
                    column_name,
                    last_per_qtr,
                    next_qtr_start_idx,
                    requested_quarter,
                    sid,
                    sid_idx,
                )
                add_new_adjustments(
                    col_to_overwrites, adjs, column_name, next_qtr_start_idx
                )
            # There are no estimates for the quarter. Overwrite all
            # values going up to the starting index of that quarter
            # with the missing value for this column.
            else:
                adjs = [self.overwrite_with_null(col, next_qtr_start_idx, sid_idx)]
                add_new_adjustments(
                    col_to_overwrites, adjs, column_name, next_qtr_start_idx
                )

    def overwrite_with_null(self, column, next_qtr_start_idx, sid_idx):
        return self.scalar_overwrites_dict[column.dtype](
            0, next_qtr_start_idx - 1, sid_idx, sid_idx, column.missing_value
        )

    def load_adjusted_array(self, domain, columns, dates, sids, mask):
        # Separate out getting the columns' datasets and the datasets'
        # num_announcements attributes to ensure that we're catching the right
        # AttributeError.
        col_to_datasets = {col: col.dataset for col in columns}
        try:
            groups = groupby(
                lambda col: col_to_datasets[col].num_announcements, col_to_datasets
            )
        except AttributeError as exc:
            raise AttributeError(
                "Datasets loaded via the "
                "EarningsEstimatesLoader must define a "
                "`num_announcements` attribute that defines "
                "how many quarters out the loader should load"
                " the data relative to `dates`."
            ) from exc
        if any(num_qtr < 0 for num_qtr in groups):
            raise ValueError(
                INVALID_NUM_QTRS_MESSAGE
                % ",".join(str(qtr) for qtr in groups if qtr < 0)
            )
        out = {}
        # To optimize performance, only work below on assets that are
        # actually in the raw data.
        data_query_cutoff_times = domain.data_query_cutoff_for_sessions(dates)
        assets_with_data = set(sids) & set(self.estimates[SID_FIELD_NAME])
        last_per_qtr, stacked_last_per_qtr = self.get_last_data_per_qtr(
            assets_with_data,
            columns,
            dates,
            data_query_cutoff_times,
        )
        # Determine which quarter is immediately next/previous for each
        # date.
        zeroth_quarter_idx = self.get_zeroth_quarter_idx(stacked_last_per_qtr)
        zero_qtr_data = stacked_last_per_qtr.loc[zeroth_quarter_idx]

        for num_announcements, columns in groups.items():
            requested_qtr_data = self.get_requested_quarter_data(
                zero_qtr_data,
                zeroth_quarter_idx,
                stacked_last_per_qtr,
                num_announcements,
                dates,
            )

            # Calculate all adjustments for the given quarter and accumulate
            # them for each column.
            col_to_adjustments = self.get_adjustments(
                zero_qtr_data, requested_qtr_data, last_per_qtr, dates, sids, columns
            )

            # Lookup the asset indexer once, this is so we can reindex
            # the assets returned into the assets requested for each column.
            # This depends on the fact that our column pd.MultiIndex has the same
            # sids for each field. This allows us to do the lookup once on
            # level 1 instead of doing the lookup each time per value in
            # level 0.
            # asset_indexer = sids.get_indexer_for(
            #     requested_qtr_data.columns.levels[1],
            # )
            for col in columns:
                column_name = self.name_map[col.name]
                # allocate the empty output with the correct missing value
                # shape = len(dates), len(sids)
                # output_array = np.full(shape=shape,
                #                        fill_value=col.missing_value,
                #                        dtype=col.dtype)
                # overwrite the missing value with values from the computed data
                try:
                    output_array = (
                        requested_qtr_data[column_name]
                        .reindex(sids, axis=1)
                        .to_numpy()
                        .astype(col.dtype)
                    )
                except Exception:
                    output_array = (
                        requested_qtr_data[column_name]
                        .reindex(sids, axis=1)
                        .to_numpy(na_value=col.missing_value)
                        .astype(col.dtype)
                    )

                # except ValueError:
                #     np.copyto(output_array[:, asset_indexer],
                #               requested_qtr_data[column_name].to_numpy(na_value=output_array.dtype),
                #               casting='unsafe')
                out[col] = AdjustedArray(
                    output_array,
                    # There may not be any adjustments at all (e.g. if
                    # len(date) == 1), so provide a default.
                    dict(col_to_adjustments.get(column_name, {})),
                    col.missing_value,
                )
        return out

    def get_last_data_per_qtr(
        self, assets_with_data, columns, dates, data_query_cutoff_times
    ):
        """Determine the last piece of information we know for each column on each
        date in the index for each sid and quarter.

        Parameters
        ----------
        assets_with_data : pd.Index
            Index of all assets that appear in the raw data given to the
            loader.
        columns : iterable of BoundColumn
            The columns that need to be loaded from the raw data.
        data_query_cutoff_times : pd.DatetimeIndex
            The calendar of dates for which data should be loaded.

        Returns
        -------
        stacked_last_per_qtr : pd.DataFrame
            A DataFrame indexed by [dates, sid, normalized_quarters] that has
            the latest information for each row of the index, sorted by event
            date.
        last_per_qtr : pd.DataFrame
            A DataFrame with columns that are a MultiIndex of [
            self.estimates.columns, normalized_quarters, sid].
        """
        # Get a DataFrame indexed by date with a MultiIndex of columns of
        # [self.estimates.columns, normalized_quarters, sid], where each cell
        # contains the latest data for that day.
        last_per_qtr = last_in_date_group(
            self.estimates,
            data_query_cutoff_times,
            assets_with_data,
            reindex=True,
            extra_groupers=[NORMALIZED_QUARTERS],
        )
        last_per_qtr.index = dates
        # Forward fill values for each quarter/sid/dataset column.
        ffill_across_cols(last_per_qtr, columns, self.name_map)
        # Stack quarter and sid into the index.
        stacked_last_per_qtr = last_per_qtr.stack(
            [SID_FIELD_NAME, NORMALIZED_QUARTERS],
        )
        # Set date index name for ease of reference
        stacked_last_per_qtr.index.set_names(
            SIMULATION_DATES,
            level=0,
            inplace=True,
        )
        stacked_last_per_qtr[EVENT_DATE_FIELD_NAME] = pd.to_datetime(
            stacked_last_per_qtr[EVENT_DATE_FIELD_NAME]
        )
        stacked_last_per_qtr = stacked_last_per_qtr.sort_values(EVENT_DATE_FIELD_NAME)
        return last_per_qtr, stacked_last_per_qtr


class NextEarningsEstimatesLoader(EarningsEstimatesLoader):
    searchsorted_side = "right"

    def create_overwrite_for_estimate(
        self,
        column,
        column_name,
        last_per_qtr,
        next_qtr_start_idx,
        requested_quarter,
        sid,
        sid_idx,
        col_to_split_adjustments=None,
        split_adjusted_asof_idx=None,
    ):
        return [
            self.array_overwrites_dict[column.dtype](
                0,
                next_qtr_start_idx - 1,
                sid_idx,
                sid_idx,
                last_per_qtr[
                    column_name,
                    requested_quarter,
                    sid,
                ].values[:next_qtr_start_idx],
            )
        ]

    def get_shifted_qtrs(self, zero_qtrs, num_announcements):
        return zero_qtrs + (num_announcements - 1)

    def get_zeroth_quarter_idx(self, stacked_last_per_qtr):
        """Filters for releases that are on or after each simulation date and
        determines the next quarter by picking out the upcoming release for
        each date in the index.

        Parameters
        ----------
        stacked_last_per_qtr : pd.DataFrame
            A DataFrame with index of calendar dates, sid, and normalized
            quarters with each row being the latest estimate for the row's
            index values, sorted by event date.

        Returns
        -------
        next_releases_per_date_index : pd.MultiIndex
            An index of calendar dates, sid, and normalized quarters, for only
            the rows that have a next event.
        """
        next_releases_per_date = (
            stacked_last_per_qtr.loc[
                stacked_last_per_qtr[EVENT_DATE_FIELD_NAME]
                >= stacked_last_per_qtr.index.get_level_values(SIMULATION_DATES)
            ]
            .groupby(
                level=[SIMULATION_DATES, SID_FIELD_NAME],
                as_index=False,
                # Here we take advantage of the fact that `stacked_last_per_qtr` is
                # sorted by event date.
            )
            .nth(0)
        )
        return next_releases_per_date.index


class PreviousEarningsEstimatesLoader(EarningsEstimatesLoader):
    searchsorted_side = "left"

    def create_overwrite_for_estimate(
        self,
        column,
        column_name,
        dates,
        next_qtr_start_idx,
        requested_quarter,
        sid,
        sid_idx,
        col_to_split_adjustments=None,
        split_adjusted_asof_idx=None,
        split_dict=None,
    ):
        return [
            self.overwrite_with_null(
                column,
                next_qtr_start_idx,
                sid_idx,
            )
        ]

    def get_shifted_qtrs(self, zero_qtrs, num_announcements):
        return zero_qtrs - (num_announcements - 1)

    def get_zeroth_quarter_idx(self, stacked_last_per_qtr):
        """Filters for releases that are on or after each simulation date and
        determines the previous quarter by picking out the most recent
        release relative to each date in the index.

        Parameters
        ----------
        stacked_last_per_qtr : pd.DataFrame
            A DataFrame with index of calendar dates, sid, and normalized
            quarters with each row being the latest estimate for the row's
            index values, sorted by event date.

        Returns
        -------
        previous_releases_per_date_index : pd.MultiIndex
            An index of calendar dates, sid, and normalized quarters, for only
            the rows that have a previous event.
        """
        previous_releases_per_date = (
            stacked_last_per_qtr.loc[
                stacked_last_per_qtr[EVENT_DATE_FIELD_NAME]
                <= stacked_last_per_qtr.index.get_level_values(SIMULATION_DATES)
            ]
            .groupby(
                level=[SIMULATION_DATES, SID_FIELD_NAME],
                as_index=False,
                # Here we take advantage of the fact that `stacked_last_per_qtr` is
                # sorted by event date.
            )
            .nth(-1)
        )
        return previous_releases_per_date.index


def validate_split_adjusted_column_specs(name_map, columns):
    to_be_split = set(columns)
    available = set(name_map.keys())
    extra = to_be_split - available
    if extra:
        raise ValueError(
            "EarningsEstimatesLoader got the following extra columns to be "
            "split-adjusted: {extra}.\n"
            "Got Columns: {to_be_split}\n"
            "Available Columns: {available}".format(
                extra=sorted(extra),
                to_be_split=sorted(to_be_split),
                available=sorted(available),
            )
        )


class SplitAdjustedEstimatesLoader(EarningsEstimatesLoader):
    """Estimates loader that loads data that needs to be split-adjusted.

    Parameters
    ----------
    split_adjustments_loader : SQLiteAdjustmentReader
        The loader to use for reading split adjustments.
    split_adjusted_column_names : iterable of str
        The column names that should be split-adjusted.
    split_adjusted_asof : pd.Timestamp
        The date that separates data into 2 halves: the first half is the set
        of dates up to and including the split_adjusted_asof date. All
        adjustments occurring during this first half are applied  to all
        dates in this first half. The second half is the set of dates after
        the split_adjusted_asof date. All adjustments occurring during this
        second half are applied sequentially as they appear in the timeline.
    """

    def __init__(
        self,
        estimates,
        name_map,
        split_adjustments_loader,
        split_adjusted_column_names,
        split_adjusted_asof,
    ):
        validate_split_adjusted_column_specs(name_map, split_adjusted_column_names)
        self._split_adjustments = split_adjustments_loader
        self._split_adjusted_column_names = split_adjusted_column_names
        self._split_adjusted_asof = split_adjusted_asof
        self._split_adjustment_dict = {}
        super(SplitAdjustedEstimatesLoader, self).__init__(estimates, name_map)

    @abstractmethod
    def collect_split_adjustments(
        self,
        adjustments_for_sid,
        requested_qtr_data,
        dates,
        sid,
        sid_idx,
        sid_estimates,
        split_adjusted_asof_idx,
        pre_adjustments,
        post_adjustments,
        requested_split_adjusted_columns,
    ):
        raise NotImplementedError("collect_split_adjustments")

    def get_adjustments_for_sid(
        self,
        group,
        dates,
        requested_qtr_data,
        last_per_qtr,
        sid_to_idx,
        columns,
        col_to_all_adjustments,
        split_adjusted_asof_idx=None,
        split_adjusted_cols_for_group=None,
    ):
        """Collects both overwrites and adjustments for a particular sid.

        Parameters
        ----------
        split_adjusted_asof_idx : int
            The integer index of the date on which the data was split-adjusted.
        split_adjusted_cols_for_group : list of str
            The names of requested columns that should also be split-adjusted.
        """
        all_adjustments_for_sid = {}
        sid = int(group.name)
        self.collect_overwrites_for_sid(
            group,
            dates,
            requested_qtr_data,
            last_per_qtr,
            sid_to_idx[sid],
            columns,
            all_adjustments_for_sid,
            sid,
        )
        (
            pre_adjustments,
            post_adjustments,
        ) = self.retrieve_split_adjustment_data_for_sid(
            dates, sid, split_adjusted_asof_idx
        )
        sid_estimates = self.estimates[self.estimates[SID_FIELD_NAME] == sid]
        # We might not have any overwrites but still have
        # adjustments, and we will need to manually add columns if
        # that is the case.
        for col_name in split_adjusted_cols_for_group:
            if col_name not in all_adjustments_for_sid:
                all_adjustments_for_sid[col_name] = {}

        self.collect_split_adjustments(
            all_adjustments_for_sid,
            requested_qtr_data,
            dates,
            sid,
            sid_to_idx[sid],
            sid_estimates,
            split_adjusted_asof_idx,
            pre_adjustments,
            post_adjustments,
            split_adjusted_cols_for_group,
        )
        self.merge_into_adjustments_for_all_sids(
            all_adjustments_for_sid, col_to_all_adjustments
        )

    def get_adjustments(
        self,
        zero_qtr_data,
        requested_qtr_data,
        last_per_qtr,
        dates,
        assets,
        columns,
        **kwargs,
    ):
        """Calculates both split adjustments and overwrites for all sids."""
        split_adjusted_cols_for_group = [
            self.name_map[col.name]
            for col in columns
            if self.name_map[col.name] in self._split_adjusted_column_names
        ]
        # Add all splits to the adjustment dict for this sid.
        split_adjusted_asof_idx = self.get_split_adjusted_asof_idx(dates)
        return super(SplitAdjustedEstimatesLoader, self).get_adjustments(
            zero_qtr_data,
            requested_qtr_data,
            last_per_qtr,
            dates,
            assets,
            columns,
            split_adjusted_cols_for_group=split_adjusted_cols_for_group,
            split_adjusted_asof_idx=split_adjusted_asof_idx,
        )

    def determine_end_idx_for_adjustment(
        self, adjustment_ts, dates, upper_bound, requested_quarter, sid_estimates
    ):
        """Determines the date until which the adjustment at the given date
        index should be applied for the given quarter.

        Parameters
        ----------
        adjustment_ts : pd.Timestamp
            The timestamp at which the adjustment occurs.
        dates : pd.DatetimeIndex
            The calendar dates over which the Pipeline is being computed.
        upper_bound : int
            The index of the upper bound in the calendar dates. This is the
            index until which the adjusment will be applied unless there is
            information for the requested quarter that comes in on or before
            that date.
        requested_quarter : float
            The quarter for which we are determining how the adjustment
            should be applied.
        sid_estimates : pd.DataFrame
            The DataFrame of estimates data for the sid for which we're
            applying the given adjustment.

        Returns
        -------
        end_idx : int
            The last index to which the adjustment should be applied for the
            given quarter/sid.
        """
        end_idx = upper_bound
        # Find the next newest kd that happens on or after
        # the date of this adjustment
        newest_kd_for_qtr = sid_estimates[
            (sid_estimates[NORMALIZED_QUARTERS] == requested_quarter)
            & (sid_estimates[TS_FIELD_NAME] >= adjustment_ts)
        ][TS_FIELD_NAME].min()
        if pd.notnull(newest_kd_for_qtr):
            newest_kd_idx = dates.searchsorted(newest_kd_for_qtr)
            # make_utc_aware(pd.DatetimeIndex(newest_kd_for_qtr))
            # We have fresh information that comes in
            # before the end of the overwrite and
            # presumably is already split-adjusted to the
            # current split. We should stop applying the
            # adjustment the day before this new
            # information comes in.
            if newest_kd_idx <= upper_bound:
                end_idx = newest_kd_idx - 1
        return end_idx

    def collect_pre_split_asof_date_adjustments(
        self,
        split_adjusted_asof_date_idx,
        sid_idx,
        pre_adjustments,
        requested_split_adjusted_columns,
    ):
        """Collect split adjustments that occur before the
        split-adjusted-asof-date. All those adjustments must first be
        UN-applied at the first date index and then re-applied on the
        appropriate dates in order to match point in time share pricing data.

        Parameters
        ----------
        split_adjusted_asof_date_idx : int
            The index in the calendar dates as-of which all data was
            split-adjusted.
        sid_idx : int
            The index of the sid for which adjustments should be collected in
            the adjusted array.
        pre_adjustments : tuple(list(float), list(int))
            The adjustment values, indexes in `dates`, and timestamps for
            adjustments that happened after the split-asof-date.
        requested_split_adjusted_columns : list of str
            The requested split adjusted columns.

        Returns
        -------
        col_to_split_adjustments : dict[str -> dict[int -> list of Adjustment]]
            The adjustments for this sid that occurred on or before the
            split-asof-date.
        """
        col_to_split_adjustments = {}
        if len(pre_adjustments[0]):
            adjustment_values, date_indexes = pre_adjustments
            for column_name in requested_split_adjusted_columns:
                col_to_split_adjustments[column_name] = {}
                # We need to undo all adjustments that happen before the
                # split_asof_date here by reversing the split ratio.
                col_to_split_adjustments[column_name][0] = [
                    Float64Multiply(
                        0,
                        split_adjusted_asof_date_idx,
                        sid_idx,
                        sid_idx,
                        1 / future_adjustment,
                    )
                    for future_adjustment in adjustment_values
                ]

                for adjustment, date_index in zip(adjustment_values, date_indexes):
                    adj = Float64Multiply(
                        0, split_adjusted_asof_date_idx, sid_idx, sid_idx, adjustment
                    )
                    add_new_adjustments(
                        col_to_split_adjustments, [adj], column_name, date_index
                    )

        return col_to_split_adjustments

    def collect_post_asof_split_adjustments(
        self,
        post_adjustments,
        requested_qtr_data,
        sid,
        sid_idx,
        sid_estimates,
        requested_split_adjusted_columns,
    ):
        """Collect split adjustments that occur after the
        split-adjusted-asof-date. Each adjustment needs to be applied to all
        dates on which knowledge for the requested quarter was older than the
        date of the adjustment.

        Parameters
        ----------
        post_adjustments : tuple(list(float), list(int), pd.DatetimeIndex)
            The adjustment values, indexes in `dates`, and timestamps for
            adjustments that happened after the split-asof-date.
        requested_qtr_data : pd.DataFrame
            The requested quarter data for each calendar date per sid.
        sid : int
            The sid for which adjustments need to be collected.
        sid_idx : int
            The index of `sid` in the adjusted array.
        sid_estimates : pd.DataFrame
            The raw estimates data for this sid.
        requested_split_adjusted_columns : list of str
            The requested split adjusted columns.
        Returns
        -------
        col_to_split_adjustments : dict[str -> dict[int -> list of Adjustment]]
            The adjustments for this sid that occurred after the
            split-asof-date.
        """
        col_to_split_adjustments = {}
        if post_adjustments:
            # Get an integer index
            requested_qtr_timeline = requested_qtr_data[SHIFTED_NORMALIZED_QTRS][
                sid
            ].reset_index()
            requested_qtr_timeline = requested_qtr_timeline[
                requested_qtr_timeline[sid].notnull()
            ]

            # Split the data into range by quarter and determine which quarter
            # was being requested in each range.
            # Split integer indexes up by quarter range
            qtr_ranges_idxs = np.split(
                requested_qtr_timeline.index,
                np.where(np.diff(requested_qtr_timeline[sid]) != 0)[0] + 1,
            )
            requested_quarters_per_range = [
                requested_qtr_timeline[sid][r[0]] for r in qtr_ranges_idxs
            ]
            # Try to apply each adjustment to each quarter range.
            for i, qtr_range in enumerate(qtr_ranges_idxs):
                for adjustment, date_index, timestamp in zip(*post_adjustments):
                    # In the default case, apply through the end of the quarter
                    upper_bound = qtr_range[-1]
                    # Find the smallest KD in estimates that is on or after the
                    # date of the given adjustment. Apply the given adjustment
                    # until that KD.
                    end_idx = self.determine_end_idx_for_adjustment(
                        timestamp,
                        requested_qtr_data.index,
                        upper_bound,
                        requested_quarters_per_range[i],
                        sid_estimates,
                    )
                    # In the default case, apply adjustment on the first day of
                    #  the quarter.
                    start_idx = qtr_range[0]
                    # If the adjustment happens during this quarter, apply the
                    # adjustment on the day it happens.
                    if date_index > start_idx:
                        start_idx = date_index
                    # We only want to apply the adjustment if we have any stale
                    # data to apply it to.
                    if qtr_range[0] <= end_idx:
                        for column_name in requested_split_adjusted_columns:
                            if column_name not in col_to_split_adjustments:
                                col_to_split_adjustments[column_name] = {}
                            adj = Float64Multiply(
                                # Always apply from first day of qtr
                                qtr_range[0],
                                end_idx,
                                sid_idx,
                                sid_idx,
                                adjustment,
                            )
                            add_new_adjustments(
                                col_to_split_adjustments, [adj], column_name, start_idx
                            )

        return col_to_split_adjustments

    def retrieve_split_adjustment_data_for_sid(
        self, dates, sid, split_adjusted_asof_idx
    ):
        """

        dates : pd.DatetimeIndex
            The calendar dates.
        sid : int
            The sid for which we want to retrieve adjustments.
        split_adjusted_asof_idx : int
            The index in `dates` as-of which the data is split adjusted.

        Returns
        -------
        pre_adjustments : tuple(list(float), list(int), pd.DatetimeIndex)
            The adjustment values and indexes in `dates` for
            adjustments that happened before the split-asof-date.
        post_adjustments : tuple(list(float), list(int), pd.DatetimeIndex)
            The adjustment values, indexes in `dates`, and timestamps for
            adjustments that happened after the split-asof-date.
        """
        adjustments = self._split_adjustments.get_adjustments_for_sid("splits", sid)
        sorted(adjustments, key=lambda adj: adj[0])
        # Get rid of any adjustments that happen outside of our date index.
        adjustments = list(filter(lambda x: dates[0] <= x[0] <= dates[-1], adjustments))
        adjustment_values = np.array([adj[1] for adj in adjustments])
        timestamps = pd.DatetimeIndex([adj[0] for adj in adjustments])
        # We need the first date on which we would have known about each
        # adjustment.
        date_indexes = dates.searchsorted(timestamps)
        pre_adjustment_idxs = np.where(date_indexes <= split_adjusted_asof_idx)[0]
        last_adjustment_split_asof_idx = -1
        if len(pre_adjustment_idxs):
            last_adjustment_split_asof_idx = pre_adjustment_idxs.max()
        pre_adjustments = (
            adjustment_values[: last_adjustment_split_asof_idx + 1],
            date_indexes[: last_adjustment_split_asof_idx + 1],
        )
        post_adjustments = (
            adjustment_values[last_adjustment_split_asof_idx + 1 :],
            date_indexes[last_adjustment_split_asof_idx + 1 :],
            timestamps[last_adjustment_split_asof_idx + 1 :],
        )
        return pre_adjustments, post_adjustments

    def _collect_adjustments(
        self,
        requested_qtr_data,
        sid,
        sid_idx,
        sid_estimates,
        split_adjusted_asof_idx,
        pre_adjustments,
        post_adjustments,
        requested_split_adjusted_columns,
    ):
        pre_adjustments_dict = self.collect_pre_split_asof_date_adjustments(
            split_adjusted_asof_idx,
            sid_idx,
            pre_adjustments,
            requested_split_adjusted_columns,
        )

        post_adjustments_dict = self.collect_post_asof_split_adjustments(
            post_adjustments,
            requested_qtr_data,
            sid,
            sid_idx,
            sid_estimates,
            requested_split_adjusted_columns,
        )
        return pre_adjustments_dict, post_adjustments_dict

    def merge_split_adjustments_with_overwrites(
        self, pre, post, overwrites, requested_split_adjusted_columns
    ):
        """Merge split adjustments with the dict containing overwrites.

        Parameters
        ----------
        pre : dict[str -> dict[int -> list]]
            The adjustments that occur before the split-adjusted-asof-date.
        post : dict[str -> dict[int -> list]]
            The adjustments that occur after the split-adjusted-asof-date.
        overwrites : dict[str -> dict[int -> list]]
            The overwrites across all time. Adjustments will be merged into
            this dictionary.
        requested_split_adjusted_columns : list of str
            List of names of split adjusted columns that are being requested.
        """
        for column_name in requested_split_adjusted_columns:
            # We can do a merge here because the timestamps in 'pre' and
            # 'post' are guaranteed to not overlap.
            if pre:
                # Either empty or contains all columns.
                for ts in pre[column_name]:
                    add_new_adjustments(
                        overwrites, pre[column_name][ts], column_name, ts
                    )
            if post:
                # Either empty or contains all columns.
                for ts in post[column_name]:
                    add_new_adjustments(
                        overwrites, post[column_name][ts], column_name, ts
                    )


class PreviousSplitAdjustedEarningsEstimatesLoader(
    SplitAdjustedEstimatesLoader, PreviousEarningsEstimatesLoader
):
    def collect_split_adjustments(
        self,
        adjustments_for_sid,
        requested_qtr_data,
        dates,
        sid,
        sid_idx,
        sid_estimates,
        split_adjusted_asof_idx,
        pre_adjustments,
        post_adjustments,
        requested_split_adjusted_columns,
    ):
        """Collect split adjustments for previous quarters and apply them to the
        given dictionary of splits for the given sid. Since overwrites just
        replace all estimates before the new quarter with NaN, we don't need to
        worry about re-applying split adjustments.

        Parameters
        ----------
        adjustments_for_sid : dict[str -> dict[int -> list]]
            The dictionary of adjustments to which splits need to be added.
            Initially it contains only overwrites.
        requested_qtr_data : pd.DataFrame
            The requested quarter data for each calendar date per sid.
        dates : pd.DatetimeIndex
            The calendar dates for which estimates data is requested.
        sid : int
            The sid for which adjustments need to be collected.
        sid_idx : int
            The index of `sid` in the adjusted array.
        sid_estimates : pd.DataFrame
            The raw estimates data for the given sid.
        split_adjusted_asof_idx : int
            The index in `dates` as-of which the data is split adjusted.
        pre_adjustments : tuple(list(float), list(int), pd.DatetimeIndex)
            The adjustment values and indexes in `dates` for
            adjustments that happened before the split-asof-date.
        post_adjustments : tuple(list(float), list(int), pd.DatetimeIndex)
            The adjustment values, indexes in `dates`, and timestamps for
            adjustments that happened after the split-asof-date.
        requested_split_adjusted_columns : list of str
            List of requested split adjusted column names.
        """
        (pre_adjustments_dict, post_adjustments_dict) = self._collect_adjustments(
            requested_qtr_data,
            sid,
            sid_idx,
            sid_estimates,
            split_adjusted_asof_idx,
            pre_adjustments,
            post_adjustments,
            requested_split_adjusted_columns,
        )
        self.merge_split_adjustments_with_overwrites(
            pre_adjustments_dict,
            post_adjustments_dict,
            adjustments_for_sid,
            requested_split_adjusted_columns,
        )


class NextSplitAdjustedEarningsEstimatesLoader(
    SplitAdjustedEstimatesLoader, NextEarningsEstimatesLoader
):
    def collect_split_adjustments(
        self,
        adjustments_for_sid,
        requested_qtr_data,
        dates,
        sid,
        sid_idx,
        sid_estimates,
        split_adjusted_asof_idx,
        pre_adjustments,
        post_adjustments,
        requested_split_adjusted_columns,
    ):
        """Collect split adjustments for future quarters. Re-apply adjustments
        that would be overwritten by overwrites. Merge split adjustments with
        overwrites into the given dictionary of splits for the given sid.

        Parameters
        ----------
        adjustments_for_sid : dict[str -> dict[int -> list]]
            The dictionary of adjustments to which splits need to be added.
            Initially it contains only overwrites.
        requested_qtr_data : pd.DataFrame
            The requested quarter data for each calendar date per sid.
        dates : pd.DatetimeIndex
            The calendar dates for which estimates data is requested.
        sid : int
            The sid for which adjustments need to be collected.
        sid_idx : int
            The index of `sid` in the adjusted array.
        sid_estimates : pd.DataFrame
            The raw estimates data for the given sid.
        split_adjusted_asof_idx : int
            The index in `dates` as-of which the data is split adjusted.
        pre_adjustments : tuple(list(float), list(int), pd.DatetimeIndex)
            The adjustment values and indexes in `dates` for
            adjustments that happened before the split-asof-date.
        post_adjustments : tuple(list(float), list(int), pd.DatetimeIndex)
            The adjustment values, indexes in `dates`, and timestamps for
            adjustments that happened after the split-asof-date.
        requested_split_adjusted_columns : list of str
            List of requested split adjusted column names.
        """
        (pre_adjustments_dict, post_adjustments_dict) = self._collect_adjustments(
            requested_qtr_data,
            sid,
            sid_idx,
            sid_estimates,
            split_adjusted_asof_idx,
            pre_adjustments,
            post_adjustments,
            requested_split_adjusted_columns,
        )
        for column_name in requested_split_adjusted_columns:
            for overwrite_ts in adjustments_for_sid[column_name]:
                # We need to cumulatively re-apply all adjustments up to the
                # split-adjusted-asof-date. We might not have any
                # pre-adjustments, so we should check for that.
                if overwrite_ts <= split_adjusted_asof_idx and pre_adjustments_dict:
                    for split_ts in pre_adjustments_dict[column_name]:
                        # The split has to have occurred during the span of
                        # the overwrite.
                        if split_ts < overwrite_ts:
                            # Create new adjustments here so that we can
                            # re-apply all applicable adjustments to ONLY
                            # the dates being overwritten.
                            adjustments_for_sid[column_name][overwrite_ts].extend(
                                [
                                    Float64Multiply(
                                        0,
                                        overwrite_ts - 1,
                                        sid_idx,
                                        sid_idx,
                                        adjustment.value,
                                    )
                                    for adjustment in pre_adjustments_dict[column_name][
                                        split_ts
                                    ]
                                ]
                            )
                # After the split-adjusted-asof-date, we need to re-apply all
                # adjustments that occur after that date and within the
                # bounds of the overwrite. They need to be applied starting
                # from the first date and until an end date. The end date is
                # the date of the newest information we get about
                # `requested_quarter` that is >= `split_ts`, or if there is no
                # new knowledge before `overwrite_ts`, then it is the date
                # before `overwrite_ts`.
                else:
                    # Overwrites happen at the first index of a new quarter,
                    # so determine here which quarter that is.
                    requested_quarter = requested_qtr_data[
                        SHIFTED_NORMALIZED_QTRS, sid
                    ].iloc[overwrite_ts]

                    for adjustment_value, date_index, timestamp in zip(
                        *post_adjustments
                    ):
                        if split_adjusted_asof_idx < date_index < overwrite_ts:
                            # Assume the entire overwrite contains stale data
                            upper_bound = overwrite_ts - 1
                            end_idx = self.determine_end_idx_for_adjustment(
                                timestamp,
                                dates,
                                upper_bound,
                                requested_quarter,
                                sid_estimates,
                            )
                            adjustments_for_sid[column_name][overwrite_ts].append(
                                Float64Multiply(
                                    0, end_idx, sid_idx, sid_idx, adjustment_value
                                )
                            )

        self.merge_split_adjustments_with_overwrites(
            pre_adjustments_dict,
            post_adjustments_dict,
            adjustments_for_sid,
            requested_split_adjusted_columns,
        )



import datetime

import pandas as pd
from exchange_calendars import ExchangeCalendar

from ziplime.assets.entities.asset import Asset
from ziplime.assets.services.asset_service import AssetService
from ziplime.data.domain.data_bundle import DataBundle
from ziplime.errors import (
    InvalidBenchmarkAsset,
    BenchmarkAssetNotAvailableTooEarly,
    BenchmarkAssetNotAvailableTooLate,
)
import polars as pl

from ziplime.exchanges.exchange import Exchange


class BenchmarkSource:
    def __init__(
            self,
            asset_service: AssetService,
            trading_calendar: ExchangeCalendar,
            sessions: pl.Series,
            exchange: Exchange,
            emission_rate: datetime.timedelta,
            benchmark_fields: frozenset[str],
            benchmark_asset: Asset | None = None,
            benchmark_returns: pl.Series | None = None,
    ):
        self.asset_service = asset_service
        self.benchmark_asset = benchmark_asset
        self.sessions = sessions
        self.emission_rate = emission_rate
        self.exchange = exchange
        self.benchmark_fields = benchmark_fields
        if len(sessions) == 0:
            self._precalculated_series = pl.Series()

        elif benchmark_asset is not None:
            self._validate_benchmark(benchmark_asset=benchmark_asset)
            self._precalculated_series = self._initialize_precalculated_series(
                asset=benchmark_asset, trading_calendar=trading_calendar, trading_days=sessions,
                exchange=exchange
            )
        elif benchmark_returns is not None:
            all_bars = pl.from_pandas(
                trading_calendar.sessions_minutes(start=sessions[0], end=sessions[-1]).tz_convert(trading_calendar.tz)
            )
            self._precalculated_series = pl.DataFrame({"date": all_bars, "close": 0.00}).group_by_dynamic(
                index_column="date", every=self.emission_rate
            ).agg(pl.col("close").sum())
        else:
            all_bars = pl.from_pandas(
                trading_calendar.sessions_minutes(start=sessions[0], end=sessions[-1]).tz_convert(trading_calendar.tz)
            )
            self._precalculated_series = pl.DataFrame({"date": all_bars, "close": 0.00}).group_by_dynamic(
                index_column="date", every=self.emission_rate
            ).agg(pl.col("close").sum())

    def get_value(self, dt: datetime.datetime) -> pd.Series:
        """Look up the returns for a given dt.

        Parameters
        ----------
        dt : datetime
            The label to look up.

        Returns
        -------
        returns : float
            The returns at the given dt or session.

        See Also
        --------
        :class:`ziplime.sources.benchmark_source.BenchmarkSource.daily_returns`

        .. warning::

           This method expects minute inputs if ``emission_rate == 'minute'``
           and session labels when ``emission_rate == 'daily``.
        """
        return self._precalculated_series.iloc[dt]

    def get_range(self, start_dt: datetime.datetime, end_dt: datetime.datetime) -> pl.DataFrame:
        """Look up the returns for a given period.

        Parameters
        ----------
        start_dt : datetime
            The inclusive start label.
        end_dt : datetime
            The inclusive end label.

        Returns
        -------
        returns : pd.Series
            The series of returns.

        See Also
        --------
        :class:`ziplime.sources.benchmark_source.BenchmarkSource.daily_returns`

        .. warning::

           This method expects minute inputs if ``emission_rate == 'minute'``
           and session labels when ``emission_rate == 'daily``.
        """
        return self._precalculated_series.filter(pl.col("date").is_between(start_dt, end_dt))

    def daily_returns(self, start: datetime.datetime, end: datetime.datetime | None = None) -> pd.Series:
        """Returns the daily returns for the given period.

        Parameters
        ----------
        start : datetime
            The inclusive starting session label.
        end : datetime, optional
            The inclusive ending session label. If not provided, treat
            ``start`` as a scalar key.

        Returns
        -------
        returns : pd.Series or float
            The returns in the given period. The index will be the trading
            calendar in the range [start, end]. If just ``start`` is provided,
            return the scalar value on that day.
        """

        # todo : returns for first day
        daily_returns = self._precalculated_series.group_by_dynamic(
            index_column="date", every="1d").agg(pl.col("close").tail(1).sum()).with_columns(
            pl.col("date").dt.date().alias("date")
        ).with_columns(pl.col("close").pct_change().alias("pct_change")).fill_null(0)

        if end is None:
            return daily_returns.filter(pl.col("date") >= start)

        return daily_returns.filter(pl.col("date").is_between(start, end))

    def _validate_benchmark(self, benchmark_asset: Asset):
        # check if this security has a stock dividend.  if so, raise an
        # error suggesting that the user pick a different asset to use
        # as benchmark.
        stock_dividends = self.asset_service.get_stock_dividends(
            sid=self.benchmark_asset.sid, trading_days=self.sessions
        )

        if len(stock_dividends) > 0:
            raise InvalidBenchmarkAsset(
                sid=str(self.benchmark_asset), dt=stock_dividends[0]["ex_date"]
            )

        if benchmark_asset.start_date > self.sessions[0]:
            # the asset started trading after the first simulation day
            raise BenchmarkAssetNotAvailableTooEarly(
                sid=str(self.benchmark_asset),
                dt=self.sessions[0],
                start_dt=benchmark_asset.start_date,
            )

        if benchmark_asset.end_date < self.sessions[-1]:
            # the asset stopped trading before the last simulation day
            raise BenchmarkAssetNotAvailableTooLate(
                sid=str(self.benchmark_asset),
                dt=self.sessions[-1],
                end_dt=benchmark_asset.end_date,
            )

    @staticmethod
    def _compute_daily_returns(g):
        return (g[-1] - g[0]) / g[0]

    @classmethod
    def downsample_minute_return_series(cls, trading_calendar: ExchangeCalendar,
                                        minutely_returns: pd.Series) -> pd.Series:
        sessions = trading_calendar.minutes_to_sessions(
            minutes=minutely_returns.index,
        )
        closes = trading_calendar.closes[sessions[0]: sessions[-1]]
        daily_returns = minutely_returns[closes].pct_change()
        daily_returns.index = closes.index
        return daily_returns.iloc[1:]

    def _initialize_precalculated_series(
            self, asset: Asset, trading_calendar: ExchangeCalendar, trading_days: pl.Series,
            exchange: Exchange
    ):
        """
        Internal method that pre-calculates the benchmark return series for
        use in the simulation.

        Parameters
        ----------
        asset:  Asset to use

        trading_calendar: TradingCalendar

        trading_days: pd.DateTimeIndex

        exchange: Exchange

        Notes
        -----
        If the benchmark asset started trading after the simulation start,
        or finished trading before the simulation end, exceptions are raised.

        If the benchmark asset started trading the same day as the simulation
        start, the first available minute price on that day is used instead
        of the previous close.

        We use history to get an adjusted price history for each day's close,
        as of the look-back date (the last day of the simulation).  Prices are
        fully adjusted for dividends, splits, and mergers.

        Returns
        -------
        returns : pd.Series
            indexed by trading day, whose values represent the %
            change from close to close.
        daily_returns : pd.Series
            the partial daily returns for each minute
        """
        all_bars: pl.Series = pl.from_pandas(
            trading_calendar.sessions_minutes(start=self.sessions[0], end=self.sessions[-1]).tz_convert(
                trading_calendar.tz)
        )  # .to_frame("date").group_by_dynamic(
        #     index_column="date", every=self.emission_rate
        # ).agg()["date"]
        limit = all_bars.to_frame("date").group_by_dynamic(
            index_column="date", every=self.emission_rate
        ).agg()["date"].len()
        # precalculated_series = pl.DataFrame({"date": all_bars, "value": 0.00}).group_by_dynamic(
        #     index_column="date", every=self.timedelta_period
        # ).agg(pl.col("value").sum())

        benchmark_series = self.exchange.get_data_by_limit(
            fields=self.benchmark_fields,
            limit=limit,
            frequency=self.emission_rate,
            end_date=all_bars[-1],
            assets=frozenset({asset}),
            include_end_date=True
        )
        return benchmark_series.with_columns(pl.col(self.benchmark_fields).pct_change().alias("pct_change"))  # [1:]



import datetime
from typing import Any

import numpy as np
import pandas as pd
import polars as pl
from exchange_calendars import ExchangeCalendar

from ziplime.data.domain.data_bundle import DataBundle
from ziplime.exchanges.exchange import Exchange
from ziplime.finance.domain.ledger import Ledger
from ziplime.finance.finance_ext import minute_annual_volatility

from ziplime.sources.benchmark_source import BenchmarkSource


class BenchmarkReturnsAndVolatility:
    """Tracks daily and cumulative returns for the benchmark as well as the
    volatility of the benchmark returns.
    """

    def start_of_simulation(
            self, ledger: Ledger, emission_rate: datetime.timedelta, trading_calendar: ExchangeCalendar,
            sessions: pd.DatetimeIndex, benchmark_source: BenchmarkSource
    ):
        daily_returns = benchmark_source.daily_returns(
            sessions[0],
            sessions[-1],
        )
        daily_returns = daily_returns.fill_nan(0.0)
        daily_returns_series = daily_returns.select("pct_change")
        if len(daily_returns_series) == 0:
            raise ValueError(f"No daily returns for benchmark. Please check if you properly loaded data for symbol {benchmark_source.benchmark_asset.get_symbol_by_exchange(exchange_name=None)} in memory.")
        self._daily_returns = daily_returns_array = daily_returns_series
        self._daily_cumulative_returns = np.cumprod(1 + daily_returns_array["pct_change"]) - 1
        self._daily_annual_volatility = (daily_returns_series.with_columns(
            expanding_sum=pl.col("pct_change").rolling_std(
                window_size=daily_returns_series.height,
                min_samples=2 if daily_returns_series.height > 1 else 1) * np.sqrt(252)
        ))["expanding_sum"]

        if emission_rate == datetime.timedelta(days=1):
            # none existing in minute mode
            self._minute_cumulative_returns = None
            self._minute_annual_volatility = None
        else:
            open_ = trading_calendar.session_open(sessions[0]).tz_convert(trading_calendar.tz).to_pydatetime()
            close = trading_calendar.session_close(sessions[-1]).tz_convert(trading_calendar.tz).to_pydatetime()
            returns = benchmark_source.get_range(start_dt=open_, end_dt=close)
            returns = returns.with_columns(
                pl.lit(0).alias("sid"),
                pl.lit(0.0).alias("close"),
                pl.lit(0).alias("pct_change"),
                pl.col("date")
            )

            # rrs = (1 + returns["pct_change"].to_pandas()).cumprod() - 1
            self._minute_cumulative_returns = (
                returns.select(pl.col("date"), pl.col("sid"), (1 + pl.col("pct_change")).cum_prod() - 1))
            min_annual_volatility = minute_annual_volatility(
                date_labels=returns["date"].dt.date(),  # returns.index.normalize().view("int64"),
                minute_returns=returns["pct_change"],
                daily_returns=daily_returns_array["pct_change"],
            )

            self._minute_annual_volatility = pl.DataFrame(
                [
                    returns.select("date").to_series(),
                    pl.Series("value", min_annual_volatility)
                ]
            )  # pl.DataFrame([pd.Series(returns.select("date"))])

    def end_of_bar(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                   exchanges: dict[str, Exchange]):
        if self._minute_cumulative_returns is None:
            # TODO: fix this so that we don't have minute/daily returns but dynamic frequency returns
            return
        r = self._minute_cumulative_returns["date" == session]["literal"][0]
        if np.isnan(r):
            r = None
        packet["cumulative_risk_metrics"]["benchmark_period_return"] = r

        v = self._minute_annual_volatility["date" == session]["value"][0]
        if np.isnan(v):
            v = None
        packet["cumulative_risk_metrics"]["benchmark_volatility"] = v

    def end_of_session(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                       exchanges: dict[str, Exchange]):
        r = self._daily_cumulative_returns[session_ix]
        if np.isnan(r):
            r = None
        packet["cumulative_risk_metrics"]["benchmark_period_return"] = r

        v = self._daily_annual_volatility[session_ix]
        if v is None or np.isnan(v):
            v = None
        packet["cumulative_risk_metrics"]["benchmark_volatility"] = v



from textwrap import dedent
from functools import partial
import numpy as np

from toolz import merge_with
from ziplime.errors import (
    WindowLengthNotPositive,
    WindowLengthTooLong,
)
from ziplime.lib.labelarray import LabelArray
from ziplime.lib.windowtemplate import AdjustedArrayWindow
from ziplime.utils.numpy_utils import (
    datetime64ns_dtype,
    float64_dtype,
    int64_dtype,
    uint8_dtype,
)
# These class names are all the same because of our bootleg templating system.


BOOL_DTYPES = frozenset(
    map(np.dtype, [np.bool_, np.uint8]),
)
FLOAT_DTYPES = frozenset(
    map(np.dtype, [np.float32, np.float64]),
)
INT_DTYPES = frozenset(
    # NOTE: uint64 not supported because it can't be safely cast to int64.
    map(np.dtype, [np.int16, np.uint16, np.int32, np.int64, np.uint32]),
)
DATETIME_DTYPES = frozenset(map(np.dtype, ["datetime64[ns]"]))
# We use object arrays for strings.
OBJECT_DTYPES = frozenset(map(np.dtype, ["O"]))
STRING_KINDS = frozenset(["S", "U"])

REPRESENTABLE_DTYPES = BOOL_DTYPES.union(
    FLOAT_DTYPES,
    INT_DTYPES,
    DATETIME_DTYPES,
    OBJECT_DTYPES,
)


def can_represent_dtype(dtype):
    """Can we build an AdjustedArray for a baseline of `dtype``?"""
    return dtype in REPRESENTABLE_DTYPES or dtype.kind in STRING_KINDS


def is_categorical(dtype):
    """Do we represent this dtype with LabelArrays rather than ndarrays?"""
    return dtype in OBJECT_DTYPES or dtype.kind in STRING_KINDS


CONCRETE_WINDOW_TYPES = {
    float64_dtype: AdjustedArrayWindow,
    int64_dtype: AdjustedArrayWindow,
    uint8_dtype: AdjustedArrayWindow,
}


def _normalize_array(data, missing_value):
    """
    Coerce buffer data for an AdjustedArray into a standard scalar
    representation, returning the coerced array and a dict of argument to pass
    to np.view to use when providing a user-facing view of the underlying data.

    - float* data is coerced to float64 with viewtype float64.
    - int32, int64, and uint32 are converted to int64 with viewtype int64.
    - datetime[*] data is coerced to int64 with a viewtype of datetime64[ns].
    - bool_ data is coerced to uint8 with a viewtype of bool_.

    Parameters
    ----------
    data : np.ndarray

    Returns
    -------
    coerced, view_kwargs : (np.ndarray, np.dtype)
        The input ``data`` array coerced to the appropriate pipeline type.
        This may return the original array or a view over the same data.
    """
    if isinstance(data, LabelArray):
        return data, {}

    data_dtype = data.dtype
    if data_dtype in BOOL_DTYPES:
        return data.astype(np.uint8, copy=False), {"dtype": np.dtype(np.bool_)}
    elif data_dtype in FLOAT_DTYPES:
        return data.astype(np.float64, copy=False), {"dtype": np.dtype(np.float64)}
    elif data_dtype in INT_DTYPES:
        return data.astype(np.int64, copy=False), {"dtype": np.dtype(np.int64)}
    elif is_categorical(data_dtype):
        if not isinstance(missing_value, LabelArray.SUPPORTED_SCALAR_TYPES):
            raise TypeError(
                "Invalid missing_value for categorical array.\n"
                "Expected None, bytes or unicode. Got %r." % missing_value,
            )
        return LabelArray(data, missing_value), {}
    elif data_dtype.kind == "M":
        try:
            outarray = data.astype("datetime64[ns]", copy=False).view("int64")
            return outarray, {"dtype": datetime64ns_dtype}
        except OverflowError as exc:
            raise ValueError(
                "AdjustedArray received a datetime array "
                "not representable as datetime64[ns].\n"
                "Min Date: %s\n"
                "Max Date: %s\n" % (data.min(), data.max())
            ) from exc
    else:
        raise TypeError(
            "Don't know how to construct AdjustedArray "
            "on data of type %s." % data_dtype
        )


def _merge_simple(adjustment_lists, front_idx, back_idx):
    """
    Merge lists of new and existing adjustments for a given index by appending
    or prepending new adjustments to existing adjustments.

    Notes
    -----
    This method is meant to be used with ``toolz.merge_with`` to merge
    adjustment mappings. In case of a collision ``adjustment_lists`` contains
    two lists, existing adjustments at index 0 and new adjustments at index 1.
    When there are no collisions, ``adjustment_lists`` contains a single list.

    Parameters
    ----------
    adjustment_lists : list[list[Adjustment]]
        List(s) of new and/or existing adjustments for a given index.
    front_idx : int
        Index of list in ``adjustment_lists`` that should be used as baseline
        in case of a collision.
    back_idx : int
        Index of list in ``adjustment_lists`` that should extend baseline list
        in case of a collision.

    Returns
    -------
    adjustments : list[Adjustment]
        List of merged adjustments for a given index.
    """
    if len(adjustment_lists) == 1:
        return list(adjustment_lists[0])
    else:
        return adjustment_lists[front_idx] + adjustment_lists[back_idx]


_merge_methods = {
    "append": partial(_merge_simple, front_idx=0, back_idx=1),
    "prepend": partial(_merge_simple, front_idx=1, back_idx=0),
}


class AdjustedArray:
    """
    An array that can be iterated with a variable-length window, and which can
    provide different views on data from different perspectives.

    Parameters
    ----------
    data : np.ndarray
        The baseline data values. This array may be mutated by
        ``traverse(..., copy=False)`` calls.
    adjustments : dict[int -> list[Adjustment]]
        A dict mapping row indices to lists of adjustments to apply when we
        reach that row.
    missing_value : object
        A value to use to fill missing data in yielded windows.
        Should be a value coercible to `data.dtype`.
    """

    __slots__ = (
        "_data",
        "_view_kwargs",
        "adjustments",
        "missing_value",
        "_invalidated",
        "__weakref__",
    )

    def __init__(self, data, adjustments, missing_value):
        self._data, self._view_kwargs = _normalize_array(data, missing_value)

        self.adjustments = adjustments
        self.missing_value = missing_value
        self._invalidated = False

    def copy(self):
        """Copy an adjusted array, deep-copying the ``data`` array."""
        if self._invalidated:
            raise ValueError("cannot copy invalidated AdjustedArray")

        return type(self)(
            self.data.copy(order="F"),
            self.adjustments,
            self.missing_value,
        )

    def update_adjustments(self, adjustments, method):
        """
        Merge ``adjustments`` with existing adjustments, handling index
        collisions according to ``method``.

        Parameters
        ----------
        adjustments : dict[int -> list[Adjustment]]
            The mapping of row indices to lists of adjustments that should be
            appended to existing adjustments.
        method : {'append', 'prepend'}
            How to handle index collisions. If 'append', new adjustments will
            be applied after previously-existing adjustments. If 'prepend', new
            adjustments will be applied before previously-existing adjustments.
        """
        try:
            merge_func = _merge_methods[method]
        except KeyError as exc:
            raise ValueError(
                "Invalid merge method %s\n"
                "Valid methods are: %s" % (method, ", ".join(_merge_methods))
            ) from exc

        self.adjustments = merge_with(
            merge_func,
            self.adjustments,
            adjustments,
        )

    @property
    def data(self):
        """
        The data stored in this array.
        """
        return self._data.view(**self._view_kwargs)

    #@lazyval
    @property
    def dtype(self):
        """
        The dtype of the data stored in this array.
        """
        return self._view_kwargs.get("dtype") or self._data.dtype

    #@lazyval
    @property
    def _iterator_type(self):
        """
        The iterator produced when `traverse` is called on this Array.
        """
        if isinstance(self._data, LabelArray):
            return AdjustedArrayWindow
        return CONCRETE_WINDOW_TYPES[self._data.dtype]

    def traverse(self, window_length, offset=0, perspective_offset=0, copy=True):
        """
        Produce an iterator rolling windows rows over our data.
        Each emitted window will have `window_length` rows.

        Parameters
        ----------
        window_length : int
            The number of rows in each emitted window.
        offset : int, optional
            Number of rows to skip before the first window.  Default is 0.
        perspective_offset : int, optional
            Number of rows past the end of the current window from which to
            "view" the underlying data.
        copy : bool, optional
            Copy the underlying data. If ``copy=False``, the adjusted array
            will be invalidated and cannot be traversed again.
        """
        if self._invalidated:
            raise ValueError("cannot traverse invalidated AdjustedArray")

        data = self._data
        if copy:
            data = data.copy(order="F")
        else:
            self._invalidated = True

        _check_window_params(data, window_length)
        return self._iterator_type(
            data,
            self._view_kwargs,
            self.adjustments,
            offset,
            window_length,
            perspective_offset,
            rounding_places=None,
        )

    def inspect(self):
        """
        Return a string representation of the data stored in this array.
        """
        return dedent(
            """\
            Adjusted Array ({dtype}):

            Data:
            {data!r}

            Adjustments:
            {adjustments}
            """
        ).format(
            dtype=self.dtype.name,
            data=self.data,
            adjustments=self.adjustments,
        )

    def update_labels(self, func):
        """
        Map a function over baseline and adjustment values in place.

        Note that the baseline data values must be a LabelArray.
        """
        if not isinstance(self.data, LabelArray):
            raise TypeError(
                "update_labels only supported if data is of type LabelArray."
            )

        # Map the baseline values.
        self._data = self._data.map(func)

        # Map each of the adjustments.
        for _, row_adjustments in self.adjustments.items():
            for adjustment in row_adjustments:
                adjustment.value = func(adjustment.value)


def ensure_adjusted_array(ndarray_or_adjusted_array, missing_value):
    if isinstance(ndarray_or_adjusted_array, AdjustedArray):
        return ndarray_or_adjusted_array
    elif isinstance(ndarray_or_adjusted_array, np.ndarray):
        return AdjustedArray(
            ndarray_or_adjusted_array,
            {},
            missing_value,
        )
    else:
        raise TypeError(
            "Can't convert %s to AdjustedArray"
            % type(ndarray_or_adjusted_array).__name__
        )


def ensure_ndarray(ndarray_or_adjusted_array):
    """
    Return the input as a numpy ndarray.

    This is a no-op if the input is already an ndarray.  If the input is an
    adjusted_array, this extracts a read-only view of its internal data buffer.

    Parameters
    ----------
    ndarray_or_adjusted_array : numpy.ndarray | ziplime.data.adjusted_array

    Returns
    -------
    out : The input, converted to an ndarray.
    """
    if isinstance(ndarray_or_adjusted_array, np.ndarray):
        return ndarray_or_adjusted_array
    elif isinstance(ndarray_or_adjusted_array, AdjustedArray):
        return ndarray_or_adjusted_array.data
    else:
        raise TypeError(
            "Can't convert %s to ndarray" % type(ndarray_or_adjusted_array).__name__
        )


def _check_window_params(data, window_length):
    """
    Check that a window of length `window_length` is well-defined on `data`.

    Parameters
    ----------
    data : np.ndarray[ndim=2]
        The array of data to check.
    window_length : int
        Length of the desired window.

    Returns
    -------
    None

    Raises
    ------
    WindowLengthNotPositive
        If window_length < 1.
    WindowLengthTooLong
        If window_length is greater than the number of rows in `data`.
    """
    if window_length < 1:
        raise WindowLengthNotPositive(window_length=window_length)

    if window_length > data.shape[0]:
        raise WindowLengthTooLong(
            nrows=data.shape[0],
            window_length=window_length,
        )



from ziplime.finance.controls.trading_control import TradingControl


class MaxOrderSize(TradingControl):
    """TradingControl representing a limit on the magnitude of any single order
    placed with the given asset.  Can be specified by share or by dollar
    value.
    """

    def __init__(self, on_error, asset=None, max_shares=None, max_notional=None):
        super(MaxOrderSize, self).__init__(
            on_error, asset=asset, max_shares=max_shares, max_notional=max_notional
        )
        self.asset = asset
        self.max_shares = max_shares
        self.max_notional = max_notional

        if max_shares is None and max_notional is None:
            raise ValueError("Must supply at least one of max_shares and max_notional")

        if max_shares and max_shares < 0:
            raise ValueError("max_shares cannot be negative.")

        if max_notional and max_notional < 0:
            raise ValueError("max_notional must be positive.")

    def validate(self, asset, amount, portfolio, algo_datetime, algo_current_data):
        """Fail if the magnitude of the given order exceeds either self.max_shares
        or self.max_notional.
        """

        if self.asset is not None and self.asset != asset:
            return

        if self.max_shares is not None and abs(amount) > self.max_shares:
            self.handle_violation(asset, amount, algo_datetime)

        current_asset_price = algo_current_data.current(asset, "price")
        order_value = amount * current_asset_price

        too_much_value = (
                self.max_notional is not None and abs(order_value) > self.max_notional
        )

        if too_much_value:
            self.handle_violation(asset, amount, algo_datetime)



import numpy as np
import pandas as pd

from ziplime.utils.date_utils import make_utc_aware
from ziplime.lib.factorize import factorize_strings

DEFAULT_FX_RATE = "DEFAULT_FX_RATE"


class FXRateReader:
    """
    Interface for reading foreign exchange (fx) rates.

    An FX rate reader contains one or more distinct "rates", each of which
    corresponds to a collection of mappings from (quote, base, dt) ->
    float. The value produced for a given (quote, base, dt) triple is the
    exchange rate to use when converting from ``base`` to ``quote`` on ``dt``.

    The specific set of rates contained in a particular reader is
    user-defined. We infer no particular semantics from their names, other than
    that they are distinct rates. Examples of possible rate names might be
    things like "bid", "mid", and "ask", or "london_close", "tokyo_close",
    "nyse_close".

    Implementations of :class:`FXRateReader` must provide at least one method::

        def get_rates(self, rate, quote, bases, dts):

    which takes a rate, a quote currency, an array of base currencies, and an
    array of dts, and produces a (len(dts), len(base))-shape array containing a
    conversion rates for all pairs in the cartesian product of bases and dts.

    Given a definition of :meth:`get_rates`, this interface automatically
    generates two additional methods::

        def get_rates_scalar(self, rate, quote, base, dt):

    and::

        def get_rates_columnar(self, rate, quote, bases, dts):

    :meth:`get_rates_scalar` takes scalar-valued ``base`` and ``dt`` values,
    and returns a scalar float value for the requested fx rate.

    :meth:`get_rates_columnar` takes parallel arrays of ``bases`` and ``dts``
    and returns a same-length array of fx rates by performing a lookup on the
    (base, dt) pairs drawn from zipping together ``bases``, and ``dts``. In
    other words, its behavior is equivalent to::

        def get_rates_columnnar(self, rate, quote, bases, dts):
            out = []
            for base, dt in zip(bases, dts):
                out.append(self.get_rate_scalar(rate, quote, base, dt))
            return np.array(out)
    """

    def get_rates(self, rate, quote, bases, dts):
        """
        Load a 2D array of fx rates.

        Parameters
        ----------
        rate : str
            Name of the rate to load.
        quote : str
            Currency code of the currency to convert into.
        bases : np.array[object]
            Array of codes of the currencies to convert from. The same currency
            may appear multiple times.
        dts : pd.DatetimeIndex
            Datetimes for which to load rates. Must be sorted in ascending
            order and localized to UTC.

        Returns
        -------
        rates : np.array
            Array of shape ``(len(dts), len(bases))`` containing foreign
            exchange rates mapping currencies from ``bases`` to ``quote``.

            The row at index i corresponds to the dt in dts[i].
            The column at index j corresponds to the base currency in bases[j].
        """

    def get_rate_scalar(self, rate, quote, base, dt):
        """
        Load a scalar FX rate value.

        Parameters
        ----------
        rate : str
            Name of the rate to load.
        quote : str
            Currency code of the currency to convert into.
        base : str
            Currency code of the currency to convert from.
        dt : np.datetime64 or pd.Timestamp
            Datetime on which to load rate.

        Returns
        -------
        rate : np.float64
            Exchange rate from base -> quote on dt.
        """
        rates_2d = self.get_rates(
            rate,
            quote,
            bases=np.array([base], dtype=object),
            dts=make_utc_aware(pd.DatetimeIndex([dt])),
        )
        return rates_2d[0, 0]

    def get_rates_columnar(self, rate, quote, bases, dts):
        """
        Load a 1D array of FX rates.

        Parameters
        ----------
        rate : str
            Name of the rate to load.
        quote : str
            Currency code of the currency to convert into.
        bases : np.array[object]
            Array of codes of the currencies to convert from. The same currency
            may appear multiple times.
        dts : np.DatetimeIndex
            Datetimes for which to load rates. The same value may appear
            multiple times. Datetimes do not need to be sorted.
        """
        if len(bases) != len(dts):
            raise ValueError(
                "len(bases) ({}) != len(dts) ({})".format(len(bases), len(dts))
            )

        bases_ix, unique_bases, _ = factorize_strings(
            bases,
            missing_value=None,
            # Only dts need to be sorted, not bases.
            sort=False,
        )
        # NOTE: np.unique returns unique_dts in sorted order, which is required
        # for calling get_rates.
        unique_dts, dts_ix = np.unique(dts.values, return_inverse=True)
        rates_2d = self.get_rates(
            rate, quote, unique_bases, pd.DatetimeIndex(unique_dts, tz="utc")
        )
        return rates_2d[dts_ix, bases_ix]



from ziplime.utils.calendar_utils import get_calendar


class ContinuousFuture:
    """Represents a specifier for a chain of future contracts, where the
    coordinates for the chain are:
    root_symbol : str
        The root symbol of the contracts.
    offset : int
        The distance from the primary chain.
        e.g. 0 specifies the primary chain, 1 the secondary, etc.
    roll_style : str
        How rolls from contract to contract should be calculated.
        Currently supports 'calendar'.

    Instances of this class are exposed to the algorithm.
    """

    # cdef readonly int64_t sid
    # # Cached hash of self.sid
    # cdef int64_t sid_hash
    #
    # cdef readonly object root_symbol
    # cdef readonly int offset
    # cdef readonly object roll_style
    #
    # cdef readonly object start_date
    # cdef readonly object end_date
    #
    # cdef readonly object exchange_info
    #
    # cdef readonly object adjustment

    _kwargnames = frozenset({
        'sid',
        'root_symbol',
        'offset',
        'start_date',
        'end_date',
        'exchange',
    })

    def __init__(self,
                 sid, # sid is required
                 root_symbol,
                 offset,
                 roll_style,
                 start_date,
                 end_date,
                 exchange_info,
                 adjustment=None):

        self.sid = sid
        self.sid_hash = hash(sid)
        self.root_symbol = root_symbol
        self.roll_style = roll_style
        self.offset = offset
        self.exchange_info = exchange_info
        self.start_date = start_date
        self.end_date = end_date
        self.adjustment = adjustment

    @property
    def exchange(self):
        return self.exchange_info.canonical_name

    @property
    def exchange_full(self):
        return self.exchange_info.name

    def __int__(self):
        return self.sid

    def __index__(self):
        return self.sid

    def __hash__(self):
        return self.sid_hash

    def __str__(self):
        return '%s(%d [%s, %s, %s, %s])' % (
            type(self).__name__,
            self.sid,
            self.root_symbol,
            self.offset,
            self.roll_style,
            self.adjustment,
        )

    def __repr__(self):
        attrs = ('root_symbol', 'offset', 'roll_style', 'adjustment')
        tuples = ((attr, repr(getattr(self, attr, None)))
                  for attr in attrs)
        strings = ('%s=%s' % (t[0], t[1]) for t in tuples)
        params = ', '.join(strings)
        return 'ContinuousFuture(%d, %s)' % (self.sid, params)

    def __reduce__(self):
        """Function used by pickle to determine how to serialize/deserialize this
        class.  Should return a tuple whose first element is self.__class__,
        and whose second element is a tuple of all the attributes that should
        be serialized/deserialized during pickling.
        """
        return (self.__class__, (self.sid,
                                 self.root_symbol,
                                 self.start_date,
                                 self.end_date,
                                 self.offset,
                                 self.roll_style,
                                 self.exchange))

    def to_dict(self):
        """Convert to a python dict."""
        return {
            'sid': self.sid,
            'root_symbol': self.root_symbol,
            'start_date': self.start_date,
            'end_date': self.end_date,
            'offset': self.offset,
            'roll_style': self.roll_style,
            'exchange': self.exchange,
        }

    @classmethod
    def from_dict(cls, dict_):
        """Build an ContinuousFuture instance from a dict."""
        return cls(**dict_)

    def is_alive_for_session(self, session_label):
        """Returns whether the continuous future is alive at the given dt.

        Parameters
        ----------
        session_label: datetime.datetime
            The desired session label to check. (midnight UTC)

        Returns
        -------
        boolean: whether the continuous is alive at the given dt.
        """
        ref_start = self.start_date.value
        ref_end = self.end_date.value

        return ref_start <= session_label.value <= ref_end

    def is_exchange_open(self, dt_minute):
        """

        Parameters
        ----------
        dt_minute: datetime.datetime (UTC, tz-aware)
            The minute to check.

        Returns
        -------
        boolean: whether the continuous futures's exchange is open at the
        given minute.
        """
        calendar = get_calendar(self.exchange)
        return calendar.is_open_on_minute(dt_minute)




import numpy as np


def check_dts(requested_dts):
    """Validate that ``requested_dts`` are valid for querying from an FX reader."""
    if not is_sorted_ascending(requested_dts):
        raise ValueError("Requested fx rates with non-ascending dts.")


def is_sorted_ascending(array):
    return (np.maximum.accumulate(array) <= array).all()



import datetime
import multiprocessing
import os
import sys
from typing import Self

import limexhub
import structlog
from asyncclick import progressbar
from joblib import Parallel, delayed

import polars as pl
from ziplime.data.services.data_bundle_source import DataBundleSource

fundamental_data_fields = [

    "total_share_holder_equity",
    "total_liabilities",
    "total_assets",
    "shares_outstanding",
    "operating_income",

    "roi",
    "roe",
    "roa",
    "revenue",
    "return_on_tangible_equity",
    "quick_ratio",
    "price_sales",
    "price_fcf",
    "pe_ratio",
    "price_book",
    "operating_margin",
    "net_worth",
    "net_income",
    "number_of_employees",
    "gross_profit",
    "eps_forecast",
    "net_profit_margin",
    "dividend_yield",
    "gross_margin",
    "ebitda",
    "eps",
    "debt_equity_ratio",
    "current_ratio",
    "long_term_debt",
]


def fetch_fundamental_data_task(date_from: datetime.datetime,
                                date_to: datetime.datetime,
                                limex_api_key: str,
                                symbol: str,
                                ) -> pl.DataFrame:
    limex_client = limexhub.RestAPI(token=limex_api_key)
    df = pl.from_pandas(limex_client.fundamental(
        symbol=symbol,
        from_date=(date_from - datetime.timedelta(days=1)).strftime("%Y-%m-%d"),
        to_date=(date_to + datetime.timedelta(days=1)).strftime("%Y-%m-%d"),
        fields=fundamental_data_fields,
    ), include_index=True)
    if len(df) > 0:
        df = df.with_columns(
            pl.lit(symbol).alias("symbol"),
            date=pl.col("date").cast(pl.Datetime).dt.replace_time_zone(str(date_from.tzinfo)),
        ).filter(pl.col("date") >= date_from, pl.col("date") <= date_to)
        return df
    return df


class LimexHubFundamentalDataSource(DataBundleSource):
    def __init__(self, limex_api_key: str, maximum_threads: int | None = None):
        super().__init__()
        self._limex_api_key = limex_api_key
        self._logger = structlog.get_logger(__name__)
        self._limex_client = limexhub.RestAPI(token=limex_api_key)
        if maximum_threads is not None:
            self._maximum_threads = min(multiprocessing.cpu_count() * 2, maximum_threads)
        else:
            self._maximum_threads = multiprocessing.cpu_count() * 2

    async def get_data(self, symbols: list[str],
                       frequency: datetime.timedelta,
                       date_from: datetime.datetime,
                       date_to: datetime.datetime,
                       **kwargs
                       ) -> pl.DataFrame:

        def fetch_fundamental_data(limex_api_key: str, symbol: str) -> pl.DataFrame | None:
            try:
                result = fetch_fundamental_data_task(date_from=date_from, date_to=date_to,
                                                     limex_api_key=limex_api_key,
                                                     symbol=symbol)
                return result
            except Exception as e:
                self._logger.exception(
                    f"Exception fetching historical data for symbol {symbol}, date_from={date_from}, date_to={date_to}. Skipping."
                )
                return None

        total_days = (date_to - date_from).days
        final = pl.DataFrame()

        with progressbar(length=len(symbols) * total_days, label="Downloading fundamental data from LimexHub",
                         file=sys.stdout) as pbar:
            res = Parallel(n_jobs=multiprocessing.cpu_count() * 2, prefer="threads",
                           return_as="generator_unordered")(
                delayed(fetch_fundamental_data)(self._limex_api_key, symbol) for symbol in symbols)
            for item in res:
                pbar.update(total_days)
                if item is None:
                    continue
                if len(item) > 0:
                    final = pl.concat([final, item])

        return final

    @classmethod
    def from_env(cls) -> Self:
        limex_hub_key = os.environ.get("LIMEX_API_KEY", None)
        maximum_threads = os.environ.get("LIMEX_HUB_MAXIMUM_THREADS", None)
        if limex_hub_key is None:
            raise ValueError("Missing LIMEX_API_KEY environment variable.")
        return cls(limex_api_key=limex_hub_key, maximum_threads=maximum_threads)



import datetime
from typing import Any

import pandas as pd
from exchange_calendars import ExchangeCalendar

from ziplime.exchanges.exchange import Exchange
from ziplime.finance.domain.ledger import Ledger

from ziplime.sources.benchmark_source import BenchmarkSource


class PNL:
    """Tracks daily and cumulative PNL."""

    def start_of_simulation(
            self, ledger: Ledger, emission_rate: datetime.timedelta, trading_calendar: ExchangeCalendar,
            sessions: pd.DatetimeIndex, benchmark_source: BenchmarkSource
    ):
        self._previous_pnl = 0.0

    def start_of_session(self, ledger, session, exchanges: dict[str, Exchange]):
        self._previous_pnl = ledger.portfolio.pnl

    def _end_of_period(self, field, packet, ledger):
        pnl = ledger.portfolio.pnl
        packet[field]["pnl"] = pnl - self._previous_pnl
        packet["cumulative_perf"]["pnl"] = ledger.portfolio.pnl

    def end_of_bar(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                   exchanges: dict[str, Exchange]):
        self._end_of_period("minute_perf", packet, ledger)

    def end_of_session(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                       exchanges: dict[str, Exchange]):
        self._end_of_period("daily_perf", packet, ledger)



"""
Position Tracking
=================

    +-----------------+----------------------------------------------------+
    | key             | value                                              |
    +=================+====================================================+
    | asset           | the asset held in this position                    |
    +-----------------+----------------------------------------------------+
    | amount          | whole number of shares in the position             |
    +-----------------+----------------------------------------------------+
    | last_sale_price | price at last sale of the asset on the exchange    |
    +-----------------+----------------------------------------------------+
    | cost_basis      | the volume weighted average price paid per share   |
    +-----------------+----------------------------------------------------+

"""
import dataclasses
import datetime

from ziplime.assets.entities.asset import Asset


@dataclasses.dataclass
class Position:
    asset: Asset
    amount: int
    cost_basis: float
    last_sale_price: float
    last_sale_date: datetime.datetime | None

    def __repr__(self):
        return f"asset: {self.asset}, amount: {self.amount}, cost_basis: {self.cost_basis}," \
               f"last_sale_price: {self.last_sale_price}"

    def to_dict(self):
        """
        Creates a dictionary representing the state of this position.
        Returns a dict object of the form:
        """
        return {
            "sid": self.asset,
            "amount": self.amount,
            "cost_basis": self.cost_basis,
            "last_sale_price": self.last_sale_price,
        }



import datetime
import math
from abc import abstractmethod

import numpy as np

from ziplime.errors import HistoryWindowStartsBeforeData
from ziplime.exchanges.exchange import Exchange
from ziplime.finance.slippage.slippage_model import SlippageModel, SQRT_252
from ziplime.finance.utils import fill_price_worse_than_limit_price
from ziplime.utils.cache import ExpiringCache


class MarketImpactBase(SlippageModel):
    """Base class for slippage models which compute a simulated price impact
    according to a history lookback.
    """

    NO_DATA_VOLATILITY_SLIPPAGE_IMPACT = 10.0 / 10000

    def __init__(self):
        super(MarketImpactBase, self).__init__()
        self._window_data_cache = ExpiringCache()

    @abstractmethod
    def get_txn_volume(self, data, order):
        """Return the number of shares we would like to order in this minute.

        Parameters
        ----------
        data : BarData
        order : Order

        Return
        ------
        int : the number of shares
        """
        raise NotImplementedError("get_txn_volume")

    @abstractmethod
    def get_simulated_impact(
            self,
            order,
            current_price,
            current_volume,
            txn_volume,
            mean_volume,
            volatility,
    ):
        """Calculate simulated price impact.

        Parameters
        ----------
        order : The order being processed.
        current_price : Current price of the asset being ordered.
        current_volume : Volume of the asset being ordered for the current bar.
        txn_volume : Number of shares/contracts being ordered.
        mean_volume : Trailing ADV of the asset.
        volatility : Annualized daily volatility of returns.

        Return
        ------
        int : impact on the current price.
        """
        raise NotImplementedError("get_simulated_impact")

    def process_order(self, exchange: Exchange, dt:datetime.datetime, order):
        if order.open_amount == 0:
            return None, None

        minute_data = data.current(order.asset, ["volume", "high", "low"])
        mean_volume, volatility = self._get_window_data(data, order.asset, 20)

        # Price to use is the average of the minute bar's open and close.
        price = np.mean([minute_data["high"], minute_data["low"]])

        volume = minute_data["volume"]
        if not volume:
            return None, None

        txn_volume = int(min(self.get_txn_volume(data, order), abs(order.open_amount)))

        # If the computed transaction volume is zero or a decimal value, 'int'
        # will round it down to zero. In that case just bail.
        if txn_volume == 0:
            return None, None

        if mean_volume == 0 or np.isnan(volatility):
            # If this is the first day the contract exists or there is no
            # volume history, default to a conservative estimate of impact.
            simulated_impact = price * self.NO_DATA_VOLATILITY_SLIPPAGE_IMPACT
        else:
            simulated_impact = self.get_simulated_impact(
                order=order,
                current_price=price,
                current_volume=volume,
                txn_volume=txn_volume,
                mean_volume=mean_volume,
                volatility=volatility,
            )

        impacted_price = price + math.copysign(simulated_impact, order.direction)

        if fill_price_worse_than_limit_price(impacted_price, order):
            return None, None

        return impacted_price, math.copysign(txn_volume, order.direction)

    def _get_window_data(self, data, asset, window_length):
        """Internal utility method to return the trailing mean volume over the
        past 'window_length' days, and volatility of close prices for a
        specific asset.

        Parameters
        ----------
        data : The BarData from which to fetch the daily windows.
        asset : The Asset whose data we are fetching.
        window_length : Number of days of history used to calculate the mean
            volume and close price volatility.

        Returns
        -------
        (mean volume, volatility)
        """
        try:
            values = self._window_data_cache.get(asset, data.current_session)
        except KeyError:
            try:
                # Add a day because we want 'window_length' complete days,
                # excluding the current day.
                volume_history = data.history(
                    asset,
                    "volume",
                    window_length + 1,
                    "1d",
                )
                close_history = data.history(
                    asset,
                    "close",
                    window_length + 1,
                    "1d",
                )
            except HistoryWindowStartsBeforeData:
                # If there is not enough data to do a full history call, return
                # values as if there was no data.
                return 0, np.nan

            # Exclude the first value of the percent change array because it is
            # always just NaN.
            close_volatility = (
                close_history[:-1]
                .pct_change()[1:]
                .std(
                    skipna=False,
                )
            )
            values = {
                "volume": volume_history[:-1].mean(),
                "close": close_volatility * SQRT_252,
            }
            self._window_data_cache.set(asset, values, data.current_session)

        return values["volume"], values["close"]



from .classifier import (
    Classifier,
    CustomClassifier,
    Quantiles,
    Everything,
    Latest,
)

__all__ = [
    "Classifier",
    "CustomClassifier",
    "Everything",
    "Latest",
    "Quantiles",
]
